Package: IONiseR
Title: Quality Assessment Tools for Oxford Nanopore MinION data
Version: 2.34.0
Authors@R: person("Mike", "Smith", email = "grimbough@gmail.com", role
        = c("aut", "cre"))
Description: IONiseR provides tools for the quality assessment of
    Oxford Nanopore MinION data. It extracts summary statistics
    from a set of fast5 files and can be used either before or
    after base calling.  In addition to standard summaries of the
    read-types produced, it provides a number of plots for
    visualising metrics relative to experiment run time or
    spatially over the surface of a flowcell.
License: MIT + file LICENSE
Depends: R (>= 3.4)
Imports: rhdf5, dplyr, magrittr, tidyr, ShortRead, Biostrings, ggplot2,
        methods, BiocGenerics, XVector, tibble, stats, BiocParallel,
        bit64, stringr, utils
VignetteBuilder: knitr
Suggests: BiocStyle, knitr, rmarkdown, gridExtra, testthat,
        minionSummaryData
biocViews: QualityControl, DataImport, Sequencing
NeedsCompilation: no
Author: Mike Smith [aut, cre]
Maintainer: Mike Smith <grimbough@gmail.com>
RoxygenNote: 6.0.1
Collate: 'IONiseR.R' 'classes.R' 'Methods-accessors.R'
        'Methods-subsetting.R' 'fast5Readers.R' 'fast5Status.R'
        'fast5readers_summary.R' 'fast5utilities.R' 'fastqProcessing.R'
        'plotting_kmers.R' 'plotting_layout.R'
        'plotting_summaryStats.R' 'processing_bam.R' 'readSummary.R'
        'squiggle.R'
git_url: https://git.bioconductor.org/packages/IONiseR
git_branch: RELEASE_3_22
git_last_commit: 25723cd
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.22
Date/Publication: 2025-10-29
Packaged: 2025-10-30 04:38:51 UTC; biocbuild

# Generated by roxygen2: do not edit by hand

export(channelActivityPlot)
export(channelHeatmap)
export(fast5toFastq)
export(layoutPlot)
export(plotActiveChannels)
export(plotBaseProductionRate)
export(plotCurrentByTime)
export(plotEventRate)
export(plotKmerFrequencyCorrelation)
export(plotReadAccumulation)
export(plotReadCategoryCounts)
export(plotReadCategoryQuals)
export(plotReadTypeProduction)
export(readFast5Log)
export(readFast5Summary)
exportClasses(Fast5Summary)
exportMethods("[")
exportMethods(baseCalled)
exportMethods(eventData)
exportMethods(fastq)
exportMethods(fastq2D)
exportMethods(fastqComplement)
exportMethods(fastqTemplate)
exportMethods(length)
exportMethods(readInfo)
import(bit64)
import(ggplot2)
import(rhdf5)
importClassesFrom(ShortRead,ShortReadQ)
importFrom(BiocGenerics,width)
importFrom(BiocParallel,MulticoreParam)
importFrom(BiocParallel,bpmapply)
importFrom(BiocParallel,register)
importFrom(Biostrings,BStringSet)
importFrom(Biostrings,DNAStringSet)
importFrom(Biostrings,oligonucleotideFrequency)
importFrom(Biostrings,quality)
importFrom(ShortRead,FastqQuality)
importFrom(ShortRead,alphabetScore)
importFrom(ShortRead,append)
importFrom(ShortRead,id)
importFrom(ShortRead,sread)
importFrom(ShortRead,strand)
importFrom(ShortRead,width)
importFrom(ShortRead,writeFastq)
importFrom(XVector,compact)
importFrom(dplyr,arrange)
importFrom(dplyr,as_data_frame)
importFrom(dplyr,count)
importFrom(dplyr,data_frame)
importFrom(dplyr,filter)
importFrom(dplyr,funs)
importFrom(dplyr,group_by)
importFrom(dplyr,inner_join)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,n)
importFrom(dplyr,order_by)
importFrom(dplyr,right_join)
importFrom(dplyr,select)
importFrom(dplyr,slice)
importFrom(dplyr,summarise)
importFrom(dplyr,summarise_each)
importFrom(dplyr,transmute)
importFrom(dplyr,with_order)
importFrom(magrittr,"%>%")
importFrom(methods,.hasSlot)
importFrom(methods,initialize)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,setClass)
importFrom(methods,setGeneric)
importFrom(methods,setMethod)
importFrom(stats,cor)
importFrom(stats,median)
importFrom(stringr,str_detect)
importFrom(stringr,str_length)
importFrom(stringr,str_match)
importFrom(stringr,str_replace)
importFrom(tibble,as_tibble)
importFrom(tibble,tibble)
importFrom(tidyr,gather)
importFrom(tidyr,spread)
importFrom(utils,packageVersion)

# IONiseR

IONiseR provides tools for the quality assessment of
    Oxford Nanopore MinION data. It extracts summary statistics
    from a set of fast5 files and can be used either before or
    after base calling.  In addition to standard summaries of the
    read-types produced, it provides a number of plots for
    visualising metrics relative to experiment run time or
    spatially over the surface of a flowcell.

## Current Status

| Travis        | BioC           | Test Coverage |
| ------------- |-------------| -----|
| [![Build Status](https://travis-ci.org/grimbough/IONiseR.svg?branch=master)](https://travis-ci.org/grimbough/IONiseR) | [![BioC Status](https://bioconductor.org/shields/build/devel/bioc/IONiseR.svg)](http://bioconductor.org/checkResults/devel/bioc-LATEST/IONiseR/) | [![Codecov](http://img.shields.io/codecov/c/github/grimbough/IONiseR.svg)](https://codecov.io/gh/grimbough/IONiseR) |

## Contact

Package Homepage: http://bioconductor.org/packages/devel/bioc/html/IONiseR.html 
Bug Reports: https://support.bioconductor.org/p/new/post/?tag_val=IONiseR.

## Funding 

Funding for continued development and maintenance of this package is provided by the German Network for Bioinformatics Infrastructure

<a href="http://www.denbi.de"><img src="https://tess.elixir-europe.org/system/content_providers/images/000/000/063/original/deNBI_Logo_rgb.jpg" width="400" align="left"></a>

#' Create layout plot of flowcell
#' 
#' Creates a plot representing the layout of a MinION flow cell.  Each circle
#' represents an individual channel with the intensity relecting a specified 
#' sequencing metric.  This function is a more generalised version of 
#' \code{\link{layoutPlot}}, allowing the user to map any value the like on 
#' the channel layout.
#' @param data A data.frame.  Should have at least two columns, 
#' one of which has the name 'channel'.
#' @param zValue Character string specifying the name of the column to be used 
#' for the colour scaling.
#' @return Returns an object of \code{gg} representing the plot.
#' @examples
#' library(dplyr)
#' if( require(minionSummaryData) ) {
#'    data(s.typhi.rep2, package = 'minionSummaryData')
#'    ## calculate and plot the mean number of events recorded by each channel
#'    avgEvents <- left_join(readInfo(s.typhi.rep2), eventData(s.typhi.rep2), by = 'id') %>% 
#'    group_by(channel) %>% 
#'    summarise(mean_nevents = mean(num_events))
#'    channelHeatmap(avgEvents, zValue = 'mean_nevents')
#' }
#' @importFrom dplyr as_tibble
#' @export
channelHeatmap <- function(data, zValue) {
    
    if(!zValue %in% colnames(data)) {
        stop("Column '", zValue, "' not found")
    }
    
    plottingMap <- as_tibble(cbind(.channelToXY(data[['channel']], shiftRows = TRUE), zValue = data[[zValue]]))
    
    ggplot(plottingMap, aes_string(x = "matrixCol", y = "matrixRow", colour = "zValue")) + 
        geom_rect(mapping = aes(xmin = 0, xmax = 8.5, ymin = -0.5, ymax = 33.5), fill = "grey50", colour = "black") +
        geom_rect(mapping = aes(xmin = 9, xmax = 17.5, ymin = -0.5, ymax = 33.5), fill = "grey50", colour = "black") +
        geom_point(size = 5.5, colour = "black") +
        geom_point(size = 4.5) + 
        scale_colour_gradient(low="darkblue", high="green") + 
        theme(panel.background = element_rect(fill = "grey50"), 
              panel.grid.major = element_line(colour = "grey50"), 
              panel.grid.minor = element_line(colour = "grey50")) + 
        scale_x_continuous(breaks=NULL) +
        scale_y_continuous(breaks=NULL) +
        xlab("") + 
        ylab("")
}

#' Create layout plot of flowcell
#' 
#' Creates a plot representing the layout of a MinION flow cell.  Each circle
#' represents an individual channel with the intensity relecting a specified 
#' sequencing metric.  This function is a more generalised version of 
#' \code{\link{layoutPlot}}, allowing the user to map any value the like on 
#' the channel layout.
#' @param data A data.frame.  Should have at least two columns, 
#' one of which has the name 'channel'.
#' @param zValue Character string specifying the name of the column to be used 
#' for the colour scaling.
#' @return Returns an object of \code{gg} representing the plot.
#' @importFrom dplyr right_join
muxHeatmap <- function(data, zValue) {
    
    if(!zValue %in% colnames(data)) {
        stop("Column '", zValue, "' not found")
    }
    
    tmp <- .muxToXY(.channelToXY(shiftRows = FALSE, insertGap = FALSE), shiftRows = TRUE)
    plottingMap <- as_tibble(right_join(tmp, data, by = c("mux" = "mux", "channel" = "channel")), zValue = data[[zValue]])
    
    tmp <- left_join(tmp, group_by(plottingMap, channel) %>% 
                         summarise(zValue = sum(zValue)), by = "channel")

    channel_data <- group_by(tmp, channel) %>% 
        summarise(row = mean(matrixRow), col = mean(matrixCol), meanZValue = mean(zValue))
    
    ggplot() +
        geom_rect(data = channel_data, 
                  mapping = aes(xmax = col+2, xmin = col-2, ymax = row+0.5, ymin = row-0.5, fill = meanZValue), color = "gray40", alpha = 1) +
        geom_polygon(data = tibble(id = rep(1:nrow(plottingMap), each = 50), 
                                       zValue = rep(plottingMap[,zValue], each = 50),
                                       rbindlist(apply(as.matrix(plottingMap[,4:3, with = FALSE]), 1, circleFun, npoints = 50, diameter = 0.8))), aes(x = x, y = y, group = id, fill = zValue), colour = "grey20") + 
        #scale_colour_gradient(low="darkblue", high="green") + 
        scale_fill_gradient(na.value = "white", low = "#999999", high = "orange") + 
        theme(panel.background = element_rect(fill = "white"), 
              panel.grid.major = element_line(colour = "white"), 
              panel.grid.minor = element_line(colour = "white")) + 
        scale_x_continuous(breaks=NULL) +
        scale_y_continuous(breaks=NULL) +
        xlab("") + 
        ylab("")
    
}

#' Create layout plot of flowcell
#' 
#' Creates a plot representing the layout of a MinION flow cell.  Each circle represents an idividual channel with the intensity relecting the total kilobases of sequence produced.  This only considers reads marked as template or complement, 2D reads are ignored as they are generated from the former two.
#' 
#' @param summaryData Object of class \linkS4class{Fast5Summary}.
#' @param attribute Character string indicating what to plot. Currently accepted values are: "nreads", "kb", "signal".
#' @return Returns an object of \code{gg} representing the plot.
#' @examples
#' if( require(minionSummaryData) ) {
#'    data(s.typhi.rep2, package = 'minionSummaryData')
#'    layoutPlot( s.typhi.rep2, attribute = 'nreads' )
#'    layoutPlot( s.typhi.rep2, attribute = 'kb' )
#' }
#' @export
#' @importFrom dplyr group_by summarise n
#' @importFrom BiocGenerics width
layoutPlot <- function(summaryData, attribute = NULL) {
    
    if(attribute == "nreads") {
        res <- group_by(readInfo(summaryData), channel) %>% 
            summarise(nreads = n())
    } else if (attribute == "kb") {
        res <- mutate(baseCalled(summaryData), 
                      seq_length = width(fastq(summaryData)[1:nrow(baseCalled(summaryData))]), 
                      channel = readInfo(summaryData)[['channel']][ match(baseCalled(summaryData)[['id']], 
                                                             readInfo(summaryData)[['id']]) ]) %>%
            group_by(channel) %>%
            summarise(kb = sum(seq_length) / 1000)  
    } else if (attribute == "signal") {
        res <- mutate(eventData(summaryData), 
                      channel = readInfo(summaryData)[match(eventData(summaryData)[,id], 
                                                            readInfo(summaryData)[,id]), channel]) %>% 
            group_by(channel) %>% 
            summarise(signal = mean(median_signal))
    } else {
        error("Unsupported attribute")
    }
    
    channelHeatmap(res, zValue = attribute) 
}


.channelToXY <- function(pore = 1:512, insertGap = TRUE, shiftRows = TRUE) {
    
    pore <- pore-1
    
    side <- (pore %/% 32) %% 2
    
    rowblock <- pore %/% 64
    rowblock[rowblock > 0] <- abs(rowblock[rowblock > 0] - 8)
    
    row <- (rowblock * 4) + ((pore %% 32) %/% 8) + 1
    column <- (-(side-1) * (pore %% 8)) + (side  * (15 - pore %% 8)) + 1
    
    if(insertGap) {
        column[column > 8] <- column[column > 8] + 1
    }
    
    if(shiftRows) {
        column[ (which(!row %% 2)) ] <- column[ (which(!row %% 2)) ] - 0.5
    }
    
    return(data.frame(channel = pore+1, matrixRow = row, matrixCol = column))
}

.muxToXY <- function(channelMap, insertGap = TRUE, shiftRows = TRUE) {
    
    muxMap <- data.frame(mux = rep(1:4, each = nrow(channelMap)),
                         rbind(channelMap, channelMap, channelMap, channelMap)) %>%
        mutate(oddEven = matrixCol %% 2)

    ##odd columns
    muxMap[mux == 1 & oddEven == 1,matrixCol:=(matrixCol*4)-1]
    muxMap[mux == 2 & oddEven == 1,matrixCol:=(matrixCol*4)]
    muxMap[mux == 3 & oddEven == 1,matrixCol:=(matrixCol*4)-3]
    muxMap[mux == 4 & oddEven == 1,matrixCol:=(matrixCol*4)-2]
    
    ##even columns
    muxMap[mux == 1 & oddEven == 0,matrixCol:=(matrixCol*4)-2]
    muxMap[mux == 2 & oddEven == 0,matrixCol:=(matrixCol*4)-3]
    muxMap[mux == 3 & oddEven == 0,matrixCol:=(matrixCol*4)]
    muxMap[mux == 4 & oddEven == 0,matrixCol:=(matrixCol*4)-1]
    
    if(insertGap) {
        muxMap[matrixCol > 32,matrixCol:=matrixCol+4]
    }
    
    if(shiftRows) {
        muxMap[as.logical(matrixRow %% 2),matrixCol:=matrixCol-0.5]
    }
    
    muxMap <- select(muxMap, -oddEven)
    muxMap
}

.circleFun <- function(center = c(0,0),diameter = 1, npoints = 100){
    r = diameter / 2
    tt <- seq(0,2*pi,length.out = npoints)
    xx <- center[1] + r * cos(tt)
    yy <- center[2] + r * sin(tt)
    return(data.frame(x = xx, y = yy))
}

## produce a ShortReadQ object from the fast5 string
#' @importFrom Biostrings BStringSet DNAStringSet
#' @importFrom ShortRead FastqQuality
#' @importFrom methods new
.processFastq <- function(string, readID = NULL, appendID = NULL) {
    
    fastqString <- strsplit(string, "\n")[[1]]
    if(is.null(readID)) {
        id <- BStringSet(paste0(fastqString[1], appendID))
    } else {
        id <- BStringSet(paste0(readIDs, appendID))
    }
    sread <- DNAStringSet(fastqString[2])
    qual <- FastqQuality(fastqString[4])
    
    new("ShortReadQ", id = id, sread = sread, quality = qual)
}

## produce a ShortReadQ object from a vector of fast5 strings
#' @importFrom Biostrings BStringSet DNAStringSet
#' @importFrom ShortRead FastqQuality
#' @importFrom BiocGenerics width
#' @importFrom stringr str_length str_match str_replace_all
.processFastqVec <- function(strings, readIDs = NULL, appendID = NULL) {
    
    if(any(str_length(strings) == 0))
        strings <- strings[-which(str_length(strings) == 0)]
    
    strings <- str_replace(string = strings, pattern = "^@", replacement = "")
    
    #fastqStrings <- strsplit(strings, "\n")
    #fastqStrings <- do.call(rbind, fastqStrings)
    fastqStrings <- stringr::str_match(strings, pattern = "(.*)\n(.*)\n(\\+)\n(.*)")[,2:5,drop=FALSE]
    
    if(is.null(readIDs)) {
        id <- BStringSet(paste0(fastqStrings[,1], appendID))
    } else {
        id <- BStringSet(paste0(readIDs, appendID))
    }
    
    ## replace U with T to cope with RNA data
    ## this could be made more generalisable
    fastqStrings[,2] <- stringr::str_replace_all(string = fastqStrings[,2], 
                             pattern = "U", 
                             replacement = "T")
    
    sread <- DNAStringSet(fastqStrings[,2])
    qual <- FastqQuality(fastqStrings[,4])
    
    ## occasionally we encounter records where 
    ## the number of bases != the number of qualities
    ## these are not valid and should be removed
    invalid <- which(width(sread) != width(qual))
    if(length(invalid)) {
        id <- id[-invalid]
        sread <- sread[-invalid]
        qual <- qual[-invalid]
    }
    
    fastq <- new("ShortReadQ", id = id, sread = sread, quality = qual)
    return(list(fastq = fastq, invalid = invalid))
}

#' @importFrom stringr str_replace
.getFastqString <- function(file, strand = "template", d = "1D",
                            dontCheck = TRUE) {
  ## returns the unprocessed fastq string stored in fast5 files
    if(is.character(file)) {
        fid <- H5Fopen(file)
        on.exit(H5Fclose(fid))
    } else {
        fid <- file
    }

    group <- paste0("/Analyses/Basecall_", d, "_000/BaseCalled_", strand, "/Fastq")
    if( dontCheck || .groupExistsObj(fid, group = group) ) {
        did <- H5Dopen(fid, group)
        fastq <- H5Dread(did)
        H5Dclose(did)
    } else {
        fastq <- ""
    }
    
    return(fastq)
}

.processStrandSpecifier <- function(strand) {
    
    strand <- stringr::str_split(strand, pattern = "\\|", simplify = TRUE)[1,]
    if(any(!strand %in% c("template", "complement", "2D", "all", "both"))) {
        stop("Unexpected option provided to 'strand'.",
             "Please provide some combination of the following options:",
             "'template', 'complement', '2D', 'all', 'both'")
    }
    
    if("all" %in% strand) {
        strand <- c("template", "complement", "2D")
    } else if ("both" %in% strand) {
        strand <- c("template", "complement")
    }
    
    return(strand)
}


#' Extract FASTQ files from fast5 files
#' 
#' This function provides direct access to the FASTQ entries held within fast5
#' files.  If you are only interested in getting hold of the base called reads,
#' and don't require any raw-signal or event information, use this function.
#' Given a vector of fast5 files, the FASTQ entries will be combined and up to
#' three gzip compressed FASTQ will be created - one for each of the template, 
#' complement and 2D strands depending upon what is available in the input 
#' files.
#' 
#' @param files Character vector of fast5 files to be read.
#' @param strand Character vector specifying the strand to extract.  Can take
#' any combination of the following options: "template", "complement", "2D", 
#' "all", "both".
#' @param fileName Stem for the name of the names of the output file names.
#' The appropriate strand will be appended to each file e.g. 
#' fileName_complement.fq.gz or fileName_template.fq.gz
#' @param outputDir Directory output files should be written to.
#' @param ncores Specify the number of CPU cores that should be used to process 
#' the files.  Currently this seems to be more IO bound than CPU, so there
#' is little benefit achieved by using a high number of cores.
#' 
#' @return No value returned.  Run for the side effect of writing the FASTQ
#' files to disk.
#' 
#' @examples \dontrun{
#' fast5files <- list.files('/foo/bar/', pattern = '.fast5$')
#' summaryData <- readFast5Summary(fast5files)
#' }
#' 
#' @export
#' @importFrom ShortRead writeFastq
#' @importFrom BiocParallel MulticoreParam bpmapply register
fast5toFastq <- function(files, strand = "all", fileName = NULL, 
                         outputDir = NULL, ncores = 1) {
    
    ## check files exist and can be accessed
    files <- files[which(file.exists(files))]
    if(length(files) == 0) {
        stop('None of the provided files can be accessed. ',
            'Have you supplied the correct path?')
    }
    
    ## understand the file structure
    status <- .fast5status(files = sample(files, size = min(length(files), 15)))
    
    strand <- .processStrandSpecifier(strand)
    ## if only 1D pipeline has been run, we can't get complement/2d data
        if( (!nchar(status$loc_complement) && "complement" %in% strand) ) {
        warning("This data has only been processed using the 1D pipeline.\n",
                "FASTQ files for the complement strand will not be generated.",
                call. = FALSE)
            strand <- strand[-which(strand == "complement")]
        }
    
    ## if only 1D pipeline has been run, we can't get complement/2d data
    if( !status$basecalled_2d && "2D" %in% strand ) {
        warning("This data does not contain 2D base calls.\n",
                "FASTQ files for the 2D strand will not be generated.",
                call. = FALSE)
        strand <- strand[-which(strand == "2D")]
    }
    
    register(MulticoreParam(workers = ncores))
    
    for(s in strand) {
        d <- str_match(status[[ paste0('loc_', s) ]], 
                       pattern = "Basecall_([12]D)")[,2]
        fastq_vec <- bpmapply(FUN = .getFastqString, files, 
                            MoreArgs = list(strand = s, d = d, dontCheck = FALSE),
                            USE.NAMES = FALSE)
        if(length(fastq_vec) == 0) {
            warning("No entries found for ", s, "strand. ",
                    "No FASTQ file created.",
                    call. = FALSE)
            next()
        }
        fastq_fq <- .processFastqVec(fastq_vec)$fastq
        writeFastq(fastq_fq, 
                   file = file.path(outputDir, 
                                    paste0(fileName, "_", s, ".fq.gz")))
    }
}

## The first read here is missing a quality value, so is invalid
testString <- c("read1\nAAAAAA\n+\n!!!!!", "read2\nCCCCCC\n+\n!!!!!!", NULL)

fastqResult <- .processFastqVec(testString)
expect_equal(length(fastqResult$invalid), 1)

#' Plot the proportion of template, complement and 2D reads found a dataset.
#' 
#' Generates a bar plot showing the breakdown of read types found in a set of fast5 files.  There is a strict hierarchy to the types of read that can be found in a fast5 file.  A full 2D read requires both a complement and template strand to have been read correctly.  Similarly, a complement strand can only be present if the template was read successfully.  Finally, you can encounter a file containing now called bases on either strand.
#' Here we visualise the total number of fast5 files, along with the counts containing each of the categories above.  For an ideal dataset all four bars will be the same height.  This is unlikely, but the drop between bars can give some indication of data quality.
#' @param summaryData Object of class \linkS4class{Fast5Summary}.
#' @return Returns an object of class \code{gg} representing the plot.
#' @examples
#' if( require(minionSummaryData) ) {
#'    data(s.typhi.rep2, package = 'minionSummaryData')
#'    plotReadCategoryCounts( s.typhi.rep2 )
#' }
#' @export
#' @importFrom dplyr summarise count
plotReadCategoryCounts <- function(summaryData) {
    tab <- c(nrow(readInfo(summaryData)),
             count(baseCalled(summaryData), strand, sort = TRUE)[['n']],
             count(baseCalled(summaryData), full_2D)[[2,'n']] / 2)
    
    if("pass" %in% names(readInfo(summaryData))) {
        pf <- any(count(readInfo(summaryData), pass)[['pass']], na.rm = TRUE)
    } else {
        pf <- FALSE
    }
    if(pf) {
        tab <- c(tab, nrow(filter(readInfo(summaryData), pass == TRUE)))
        res <- tibble(
            category = factor(c('Fast5 File Count', 'Template', 'Complement', 'Full 2D', 'Pass'),
                              levels = c('Fast5 File Count', 'Template', 'Complement', 'Full 2D', 'Pass')),   
            count = tab)
    } else {
        res <- tibble(
            category = factor(c('Fast5 File Count', 'Template', 'Complement', 'Full 2D'),
                              levels = c('Fast5 File Count', 'Template', 'Complement', 'Full 2D')),   
            count = tab)
    }

    ggplot(res, aes(x = category, y = count, fill = category)) + 
        geom_bar(stat = 'identity') +
        xlab('read type') +
        guides(fill=FALSE)
}

#' Visualise the mean base quality of each read
#' 
#' Generates a box plot showing the mean base quality for each read, broken down into the three categories of read type that can be found in a fast5 file.
#' 
#' @param summaryData Object of class \linkS4class{Fast5Summary}.
#' @return Returns an object of class \code{gg} representing the plot.
#' @examples
#' if( require(minionSummaryData) ) {
#'    data(s.typhi.rep2, package = 'minionSummaryData')
#'    plotReadCategoryQuals( s.typhi.rep2 )
#' }
#' @export
#' @importFrom ShortRead alphabetScore
#' @importFrom Biostrings quality
plotReadCategoryQuals <- function(summaryData) {
    
    fq <- fastq(summaryData)
    ## i've inserted an extra level in the factor here as I like the 4 colour option more - 
    ## probaby not the best thing to do.
    readType <- factor(.readtypeFromFASTQ(fq), levels = c('space', 'template', 'complement', '2D'))
    meanBaseQuality <- alphabetScore(quality(fq)) / width(fq) 
    
    res <- data.frame(readType, meanBaseQuality)
    ggplot(res, aes(x = readType, y = meanBaseQuality, fill = factor(readType))) + 
        geom_boxplot() +
        xlab('read type') +
        ylab('mean base quality per read') +
        guides(fill=FALSE)
}

#' Plot the number of active channels for each minute of run time
#' @param summaryData Object of class \linkS4class{Fast5Summary}.
#' @return Returns an object of class \code{gg} representing the plot.
#' @examples
#' if( require(minionSummaryData) ) {
#'    data(s.typhi.rep2, package = 'minionSummaryData')
#'    plotActiveChannels( s.typhi.rep2 )
#' }
#' @export
#' @importFrom dplyr mutate count tibble
plotActiveChannels <- function(summaryData) {
    startEndSummary <- mutate(eventData(summaryData), first = start_time %/% 60, last = (start_time + duration) %/% 60)
    tab <- tibble(minute = unlist(apply(startEndSummary, 1, function(x) { x['first']:x['last'] }))) %>%
        count(minute)
    
    ggplot(tab, aes(x = minute / 60, y = n)) + 
        geom_point() + 
        xlab("hour") +
        ylab("active channels")
}

#' Plot the accumulation of reads over the duration of the experiment.
#' @param summaryData Object of class \linkS4class{Fast5Summary}.
#' @return Returns an object of class \code{gg} representing the plot.
#' @examples
#' if( require(minionSummaryData) ) {
#'    data(s.typhi.rep2, package = 'minionSummaryData')
#'    plotReadAccumulation( s.typhi.rep2 )
#' }
#' @export
#' @importFrom dplyr group_by summarise mutate order_by with_order n
plotReadAccumulation <- function(summaryData) {
    readAccumulation <- group_by(eventData(summaryData), minute = start_time %/% 60) %>%
        summarise(new_reads = n()) %>%
        mutate(accumulation = order_by(minute, cumsum(new_reads)))
    ggplot(readAccumulation, aes(x = minute / 60, y = accumulation)) + 
        geom_point() + 
        xlab("hour") +
        ylab("reads produced")
}


#' Plot the mean rate at which events occur 
#' 
#' For each read, the ratio between the number of events comprising the read and the time spent in the pore is calculated.  This is then plotted against the time the read entered the pore, allow us to assess whether the rate at which events occur changes during the experiment run time.
#' @param summaryData Object of class \linkS4class{Fast5Summary}.
#' @return Returns an object of class \code{gg} representing the plot.
#' @examples
#' if( require(minionSummaryData) ) {
#'    data(s.typhi.rep2, package = 'minionSummaryData')
#'    plotEventRate( s.typhi.rep2 )
#' }
#' @export
plotEventRate <- function(summaryData) {
    ggplot(eventData(summaryData), aes(x = start_time %/% 60, y = num_events / duration)) + 
        geom_point(alpha = 0.3) + 
        ylim(0,80) +
        xlab("minute")
}

#' Plot the mean rate at which bases are recorded 
#' 
#' For each read, the ratio between the total number of bases called in the read
#' (template and complement strand, but not 2D composite) and the time spent in
#' the pore is calculated.  This is then plotted against the time the read 
#' entered the pore, allow us to assess whether the rate at which callable 
#' bases are read changes during the experiment run time.
#' 
#' This is likely very similar to \code{\link{plotEventRate}}, although one may
#' find that large number of events occur that can not be base called, 
#' resulting in a difference between these two plots.
#' 
#' 
#' @param summaryData Object of class \linkS4class{Fast5Summary}.
#' @return Returns an object of class \code{gg} representing the plot.
#' @examples
#' if( require(minionSummaryData) ) {
#'    data(s.typhi.rep2, package = 'minionSummaryData')
#'    plotBaseProductionRate( s.typhi.rep2 )
#' }
#' @export
#' @importFrom dplyr mutate 
#' @importFrom ShortRead width
plotBaseProductionRate <- function(summaryData) {
    ## select only the fastq records for the template and complement reads
    ## ignore the composite 2D reads here
    recordTable <- .matchRecords(summaryData)
    fastqIDX <- c(recordTable[['fastqTemplate']], recordTable[['fastqComplement']])
    fastqIDX <- fastqIDX[-which(is.na(fastqIDX))]
    res <- mutate(baseCalled(summaryData), bases_called = width(fastq(summaryData)[ fastqIDX ]))
    ggplot(res, aes(x = start_time %/% 60, y = bases_called / duration)) + 
        geom_point(alpha = 0.3) + 
        ylim(0,80) +
        xlab("minute")
}

#' View changes in signal against run time.
#' 
#' Plots the median recorded current for each fast5 file against the time at 
#' which the recording began.
#' 
#' @param summaryData Object of class \linkS4class{Fast5Summary}.
#' @return Returns an object of class \code{gg} representing the plot.
#' @examples
#' if( require(minionSummaryData) ) {
#'    data(s.typhi.rep2, package = 'minionSummaryData')
#'    plotCurrentByTime( s.typhi.rep2 )
#' }
#' @export
plotCurrentByTime <- function(summaryData) {
    ggplot(eventData(summaryData), aes(x = start_time / 3600, y = median_signal)) + 
        geom_point() + 
        xlab("hour") +
        ylab("median current (pA)")
}


#' View changes in signal against run time.
#' 
#' Plots the median recorded current for each fast5 file against the time at 
#' which the recording began.
#' 
#' @param summaryData Object of class \linkS4class{Fast5Summary}.
#' @param groupedMinutes Integer specifying how many minutes of runtime should be grouped together.
#' @return Returns an object of class \code{gg} representing the plot.
#' @examples
#' if( require(minionSummaryData) ) {
#'    data(s.typhi.rep2, package = 'minionSummaryData')
#'    plotReadTypeProduction( s.typhi.rep2 )
#' }
#' @export
#' @importFrom dplyr left_join
plotReadTypeProduction <- function(summaryData, groupedMinutes = 10) {
    tmp <- left_join(baseCalled(summaryData), readInfo(summaryData), by = 'id') %>%
        filter(strand == "template") %>%
        mutate(time_group = start_time %/% (60 * groupedMinutes)) %>%
        group_by(time_group, full_2D, pass) %>%
        summarise(count = n()) %>%
        mutate(hour = (time_group * groupedMinutes)/60 )
    
    ggplot(tmp, aes(x = hour, y = count, colour = interaction(full_2D, pass))) + 
        geom_point(size = 3) +
    scale_colour_manual(values=c("#999999", "#E69F00", "#56B4E9"), 
                      name="Category",
                      labels=c("Not 2D", "2D - Fail", "2D - Pass"))
}

#' @importFrom dplyr inner_join
plot2DYield <- function(summaryData, groupedMinutes = 1) {
    
    only2d <- .get2D(summaryData)
    tmp.fq <- fastq(only2d)[grep("2D", id(fastq(only2d))),]
    
    tmp <- inner_join(readInfo(only2d), tibble(id = .idFromFASTQ(tmp.fq), nbases = width(tmp.fq)), by = "id")
    tmp <- inner_join(tmp, baseCalled(only2d), by = "id")
    
    readAccumulation <- group_by(tmp, time_group = start_time %/% (60 * groupedMinutes), pass) %>%
        summarise(nbases = sum(nbases), hour = (time_group * groupedMinutes)/60 ) %>%
        arrange(time_group) %>%
        group_by(pass) %>%
        mutate(accumulation = order_by(time_group, cumsum(nbases)))
    ggplot(readAccumulation, aes(x = hour, y = accumulation, colour = pass)) + 
        geom_point() + 
        ylab("bases produced")
}

#' Visualise a specified metric over all channels over time.
#' 
#' Plots a line for each fast 5 file, arranged by channel and experiment time 
#' when the signal was being recorded.  The colour of each line can be 
#' specified by the user to reflect any metric they wish.  The intention of the
#' plot is to investigate trends that may appear at specific time points, or 
#' influence a subset of channels.
#' 
#' @param summaryData Object of class \linkS4class{Fast5Summary}.
#' @param zScale A data.frame containg two columns.  The first must be labelled
#' 'id' and correspond to id field present in all slots in summaryData.  The 
#' second column should contain data pertaining to that reads that you wish to
#' be represented on the coloured z-axis.
#' @param zAverage Logical indicating if a bar showing the mean across all 
#' channel for the choosen zScale should be shown on the plot. Defaults to
#' TRUE.
#' @return Returns an object of class \code{gg} representing the plot.
#' @examples
#' if( require(minionSummaryData) ) {
#'    require(dplyr)
#'    data(s.typhi.rep3, package = 'minionSummaryData')
#'    ## we will plot the median event signal for each read on z-axis
#'    z_scale = select(eventData(s.typhi.rep3), id, median_signal)
#'    channelActivityPlot( s.typhi.rep3, zScale = z_scale )
#' }
#' @export
#' @importFrom dplyr left_join select
channelActivityPlot <- function(summaryData, zScale = NULL, zAverage = TRUE) {
    
    tmp <- left_join(readInfo(summaryData), eventData(summaryData), by = 'id') %>%
        select(id, channel, start_time, duration)
    
    ## if we've provided a zvalue, add it to our table and set the column name
    if(!is.null(zScale)) {
        names(zScale)[ncol(zScale)] <- "zvalue"
        tmp <- left_join(tmp, zScale, by = "id")
    } else {
        tmp <- mutate(tmp, zvalue = 1)
    }
    
    p1 <- ggplot(NULL) +
        geom_segment(data = tmp,
                     mapping = aes(y = channel, yend = channel, 
                                   x = (start_time / 3600), xend = (start_time + duration) / 3600,
                                   color = zvalue), 
                     size = 1) +
        xlab("hour") +
        scale_x_continuous(expand = c(0, 0)) 
    
    if(zAverage) {
        tmp2 <- tmp %>% 
            group_by(time_bin = start_time %/% 600) %>%
            summarise(mean_value = mean(zvalue, na.rm = TRUE))
        p1 <- p1 + geom_rect(mutate(tmp2, start_time = (time_bin * 600) / 3600), mapping = aes(xmin = start_time, xmax = start_time + (600/3600), ymin = -100, ymax = -25, fill = mean_value)) +
            scale_fill_gradient(low = "darkblue", high = "orange", name = "Average\nacross\nchannels")
        
    }
        
    p1 + scale_color_gradient(low = "darkblue", high = "orange", name = "Individual\nreads")
}

library(IONiseR)
context('Testing plotting functions')

require(minionSummaryData)
data(s.typhi.rep2, package = "minionSummaryData")

p1 <- plotReadCategoryCounts(s.typhi.rep2)
test_that("plot is ggplot", {
    expect_is(p1, "ggplot")
})

test_that("category counts identified", {
    expect_equal(p1$data[['count']], c(3591, 3088, 1678, 1412, 1019))
    expect_equal(p1$data[['category']], 
                 factor(c("Fast5 File Count", "Template", "Complement", "Full 2D", "Pass"),
                        levels = c("Fast5 File Count", "Template", "Complement", "Full 2D", "Pass")))
})

p2 <- plotActiveChannels(s.typhi.rep2)
test_that("plot is ggplot", {
    expect_is(p2, "ggplot")
})

p3 <- plotReadAccumulation(s.typhi.rep2)
test_that("plot is ggplot", {
    expect_is(p3, "ggplot")
})

p4 <- plotReadTypeProduction(s.typhi.rep2)
test_that("plot is ggplot", {
    expect_is(p4, "ggplot")
})

p5 <- plotKmerFrequencyCorrelation(s.typhi.rep2)
test_that("plot is ggplot", {
    expect_is(p5, "ggplot")
})

.getRunID <- function(file) {
    
    fid <- H5Fopen(file)
    on.exit(H5Fclose(fid))
    
    exists <- .groupExistsObj(fid, group = "/UniqueGlobalKey/tracking_id")
    if(!exists) {
        run_id <- NA
    } else {
        gid <- H5Gopen(fid, "/UniqueGlobalKey/tracking_id")   
        aid <- H5Aopen(gid, "run_id")
        run_id <- H5Aread(aid)
        H5Aclose(aid)
        H5Gclose(gid)
    }
    return( run_id )
    
}

.getStartTime <- function(file, samplingRate = NULL, readNum = NULL) {
    
    if(is.character(file)) {
        fid <- H5Fopen(file)
        on.exit(H5Fclose(fid))
    } else {
        fid <- file
    }
    
    start_time <- NA
    
    ## providing the read number might speed things up if we have it already
    if(is.null(readNum)) {
        gid <- H5Gopen(fid, "/Analyses/EventDetection_000/Reads")
        readNum <- h5ls(gid)[1,"name"]
        H5Gclose(gid)
    }
    
    if(.groupExistsObj(fid, group = paste0("/Raw/EventDetection_000/Reads/", readNum))) {
        gid <- H5Gopen(fid, paste0("/Analyses/EventDetection_000/Reads/", readNum)) 
        aid <- H5Aopen(gid, "start_time") 
        start_time <- H5Aread(aid) 
        H5Aclose(aid)
    }
    
    ## start time should be stored under /Raw/Reads, so try here too
    ## for old files this doesn't exist, and sometimes start_time is blank
    if(is.na(start_time) && 
       .groupExistsObj(fid, group = paste0("/Raw/Reads/", readNum))) {
        gid <- H5Gopen(fid, paste0("/Raw/Reads/", readNum)) 
        aid <- H5Aopen(gid, "start_time") 
        start_time <- H5Aread(aid) 
        H5Aclose(aid)
    }
    
    ## start_time is recorded in number of samples. If we have to sampling rate
    ## we can convert this to seconds
    if(!is.null(samplingRate)) {
        start_time <- start_time / samplingRate
    }
    
    return(start_time)
} 

## We can often get temporal information from either event data
## or raw data.  This function gets the start time and durtion
## from the raw data group in the fast5 file
.getRawStartDuration <- function(file, readNum = NULL) {
    
    if(is.character(file)) {
        fid <- H5Fopen(file)
        on.exit(H5Fclose(fid))
    } else {
        fid <- file
    }
    
    start_time <- NA
    
    ## providing the read number might speed things up if we have it already
    ## wont work for now, later code expects just the number not 'Read_N'
    #if(is.null(readNum)) {
    #    gid <- H5Gopen(fid, "/Raw/Reads/")
    #    readNum <- h5ls(gid)[1,"name"]
    #    H5Gclose(gid)
    #}
    
    ## start time should be stored under /Raw/Reads, so try here too
    ## for old files this doesn't exist, and sometimes start_time is blank
    if(is.na(start_time) && 
       .groupExistsObj(fid, group = paste0("/Raw/Reads/Read_", readNum))) {
        gid <- H5Gopen(fid, paste0("/Raw/Reads/Read_", readNum)) 
        aid <- H5Aopen(gid, "start_time") 
        start_time <- H5Aread(aid) 
        H5Aclose(aid)
        
        aid <- H5Aopen(gid, "duration") 
        duration <- H5Aread(aid) 
        H5Aclose(aid)
        H5Gclose(gid)
    }
    
    return( c(start_time, duration) )
} 

## The sampling is how many times the signal is recorded per second.
## We use this to convert the 'duration' and 'start_time' in the event data
## into seconds.  It may also be useful meta data
#' @importFrom stats median
.getSamplingRate <- function(file) {
    
    fid <- H5Fopen(file)
    on.exit(H5Fclose(fid))
    
    exists <- .groupExistsObj(fid, group = "/UniqueGlobalKey/channel_id")
    if(!exists) {
        rate <- NA
    } else {
        gid <- H5Gopen(fid, "/UniqueGlobalKey/channel_id/")   
        aid <- H5Aopen(gid, "sampling_rate")
        rate <- H5Aread(aid)
        H5Aclose(aid)
        H5Gclose(gid)
    }
    return( rate )
}


.getReadNumber <- function(fid) {
    gid <- H5Gopen(fid, "/Analyses/EventDetection_000/Reads")
    on.exit(H5Gclose(gid))
    read_number_char <- h5ls(gid)[1,"name"]
    read_number_char <- stringr::str_replace(string = read_number_char, pattern = "Read_", replacement = "")
    read_number <- as.integer(read_number_char)
    return(read_number)
}

.getReadChannelMux <- function(file) {
    
    if(is.character(file)) {
        fid <- H5Fopen(file)
        on.exit(H5Fclose(fid))
    } else {
        fid <- file
    }
    
    # here we get the starting mux and the read number from the channel
    exists <- .groupExistsObj(fid, group = "/Analyses/EventDetection_000/Reads")
    if(!exists) {
        start_mux <- NA
        read_number <- NA
    } else {
        ## get the Read_No., this changes in every file
        gid <- H5Gopen(fid, "/Analyses/EventDetection_000/Reads")
        read_number_char <- h5ls(gid)[1,"name"]
        H5Gclose(gid)
        
        ## Open the group and read the two attribute we want
        gid <- H5Gopen(fid, paste0("/Analyses/EventDetection_000/Reads/", read_number_char))   
        aid <- H5Aopen(gid, "start_mux")
        start_mux <- H5Aread(aid)
        H5Aclose(aid)
        aid <- H5Aopen(gid, "read_number")
        read_number <- H5Aread(aid)
        H5Aclose(aid)   
        H5Gclose(gid)
    }
    
    # we're also interested in the channel information
    exists <- .groupExistsObj(fid, group = "/UniqueGlobalKey/channel_id")
    if(!exists) {
        channel_number <- NA
    } else {
        gid <- H5Gopen(fid, "/UniqueGlobalKey/channel_id/")   
        aid <- H5Aopen(gid, "channel_number")
        channel_number <- H5Aread(aid)
        H5Aclose(aid)
        H5Gclose(gid)
    }
    
    return( tibble(read = as.integer(read_number), 
                   channel = as.integer(channel_number), 
                   mux = as.integer(start_mux)) ) 
}

.getReadChannelMux2 <- function(file, readNumber = NA, dontCheck = FALSE) {
    
    if(is.character(file)) {
        fid <- H5Fopen(file)
        on.exit(H5Fclose(fid))
    } else {
        fid <- file
    }
    
    # here we get the starting mux and the read number from the channel
    #exists <- .groupExistsObj(fid, group = "/Analyses/EventDetection_000/Reads")
    if(dontCheck || .groupExistsObj(fid, group = "/Analyses/EventDetection_000/Reads")) {
        ## get the Read_No., this changes in every file
        if(is.na(readNumber)) {
            readNumber <- .getReadNumber(fid)
        }
        
        ## Open the group and read the two attribute we want
        gid <- H5Gopen(fid, paste0("/Analyses/EventDetection_000/Reads/Read_", readNumber))   
        aid <- H5Aopen(gid, "start_mux")
        start_mux <- H5Aread(aid)
        H5Aclose(aid)
        H5Gclose(gid)
    } else {
        start_mux <- NA
    }
    
    # we're also interested in the channel information
    if(dontCheck || .groupExistsObj(fid, group = "/UniqueGlobalKey/channel_id")) {
        gid <- H5Gopen(fid, "/UniqueGlobalKey/channel_id/")   
        aid <- H5Aopen(gid, "channel_number")
        channel_number <- H5Aread(aid)
        H5Aclose(aid)
        H5Gclose(gid)
    } else {
        channel_number <- NA
    }
    
    return( tibble(read = as.integer(readNumber), 
                   channel = as.integer(channel_number), 
                   mux = as.integer(start_mux)) ) 
}


.getEvents <- function(file) {
    
    fid <- H5Fopen(file)
    on.exit(H5Fclose(fid))
    
    exists <- .groupExistsObj(fid, group = "/Analyses/EventDetection_000/Reads")
    if(!exists) {
        start_time <- duration <- num_events <- median_signal <- NA
    } else {
        ## get the Read_No., this changes in every file
        gid <- H5Gopen(fid, "/Analyses/EventDetection_000/Reads")
        read_number_char <- h5ls(gid)[1,"name"]
        H5Gclose(gid)
        
        ## Open the group
        gid <- H5Gopen(fid, paste0("/Analyses/EventDetection_000/Reads/", read_number_char)) 
        did <- H5Dopen(gid, "Events")
        events <- as_tibble(H5Dread(did, bit64conversion = "int", compoundAsDataFrame = TRUE))
        H5Dclose(did)
        
        H5Gclose(gid)
    } 
    
    return(events)
}

## we want to determine if there are more than one Basecalled slots in the 
## fast5 file e.g. Basecalled_1D_000 and Basecalled_1D_001 etc.  We will use
## the highest number to read from
#' @importFrom stringr str_match
.findAnalysisNumber <- function(fid, grepString = "Basecall_[12]D_([0-9]+)") {
    
    groups <- unique(h5ls(fid)[,'group'])
    options <- unique(str_match(string = groups, pattern = grepString)[,2])
    return( options[ which.max(as.integer(options)) ] )
}



.getBaseCalled <- function(file, strand = "template") {
    
    fid <- H5Fopen(file)
    on.exit(H5Fclose(fid))
    
    analysisNum <- .findAnalysisNumber(fid)
    
    for(d in c("1D", "2D")) {
        exists <- .groupExistsObj(fid, group = paste0("/Analyses/Basecall_", d,
                                                      "_", analysisNum, 
                                                      "/Summary/basecall_1d_", strand))
        if(exists) break;
    }
    
    if(!exists) {
        events <- NA
    } else {
        ## Open the group and read
        did <- H5Dopen(fid, paste0("/Analyses/Basecall_", d, "_", analysisNum, "/BaseCalled_", strand, "/Events"))   
        events <- as_tibble(H5Dread(did, bit64conversion = "int", compoundAsDataFrame = TRUE))
        H5Dclose(did)
    }

    return(events)  
}

.getRawSignal <- function(file) {

    if(is.character(file)) {
        fid <- H5Fopen(file)
        on.exit(H5Fclose(fid))
    } else {
        fid <- file
    }

    exists <- .groupExistsObj(fid, group = "/Raw/Reads")
    if(!exists) {
        #warning("Raw signal data not found")
        signal <- NA
    } else {
        ## get the Read_No., this changes in every file
        gid <- H5Gopen(fid, "/Raw/Reads")
        read_number_char <- h5ls(gid)[1,"name"]
        H5Gclose(gid)

        ## Open the group
        gid <- H5Gopen(fid, paste0("/Raw/Reads/", read_number_char))
        
        did <- H5Dopen(gid, "Signal")
        signal = H5Dread(did)
        H5Dclose(did)

        ## get the starting time
        aid <- H5Aopen(gid, name = "start_time")
        startTime <- H5Aread(aid)
        H5Aclose(aid)
        
        H5Gclose(gid)

        raw <- tibble(signal, time = seq(startTime, length.out = length(signal)))
    }

    return(raw)
}

.getAligned <- function(file) {
 
    fid <- H5Fopen(file)
    on.exit(H5Fclose(fid))
    
    exists <- .groupExistsObj(fid, group = "/Analyses/Alignment_000/Aligned_2d/SAM")
    
    if(exists) {
        did <- H5Dopen(fid, "/Analyses/Alignment_000/Aligned_2d/SAM")
        samData <- tibble(H5Dread(did))
    } else {
        samData <- ""
    }
    
    return(samData)
}



#' Read the log entry from a single fast5 file
#' 
#' Basecalling procedures performed on fast5 files generally leave a 
#' log file entry recording how far through the pipeline the file 
#' proceeded.  This function will extract this information as a 
#' single string.  It can be printed in a more readable format 
#' using the \code{\link{cat}} function.
#' 
#' @param file Character vector of fast5 file to be read.
#' @return Character vector containing the log information.  
#' \code{NULL} if no log is found.
#' @examples 
#' fast5file <- system.file('extdata', 'example.fast5', package = "IONiseR")
#' log <- readFast5Log(fast5file)
#' cat(log)
#' @export
readFast5Log <- function(file) {

    fid <- H5Fopen(file)
    on.exit(H5Fclose(fid))
    
    log <- NULL
    
    for(d in c("1D", "2D")) {
        path <- paste0("/Analyses/Basecall_", d, "_000/Log")
        exists <- .groupExistsObj(fid, group = path)
        if(exists) {
            did <- H5Dopen(fid, path)
            log <- c(log, H5Dread(did))
        }
    }
    
    if(is.null(log))
        message("No log information found.")
    
    return(log)
}

## bamProcessing

# @importFrom stringr str_locate_all str_match
# @importFrom Rsamtools asBam
#.createBamFile <- function(strings, bamFile = NULL) {
    
    ## we assume all reads have been aligned to the same reference, so use the header from the first one
#    nl.idx <- str_locate_all(strings[1], "\n")[[1]][,1]
#    samHeader <- str_sub(strings[1], start = 1, end = sort(nl.idx, decreasing = TRUE)[2]-1)
    
    ## find the entry for the read, ignore the header
#    samRecords <- str_match(strings, "\n([[:print:]\t]*)\n$")[,2]
    
#    tmpFile <- tempfile()
#    writeLines(c(samHeader, samRecords), con = tmpFile)
#    asBam(file = tmpFile, destination = bamFile)
#}

## ----style, echo = FALSE, results = 'asis'------------------------------------
BiocStyle::markdown()
knitr::opts_chunk$set(eval=TRUE)

## ----loadLibs, message=FALSE, warning=FALSE-----------------------------------
library(IONiseR)
library(ggplot2)
library(gridExtra)

## ----readData, eval=FALSE-----------------------------------------------------
# fast5files <- list.files(path = "/path/to/data/", pattern = ".fast5$", full.names = TRUE)
# example.summary <- readFast5Summary( fast5files )

## ----loadData-----------------------------------------------------------------
library(minionSummaryData)
data(s.typhi.rep1)

## ----exampleData--------------------------------------------------------------
s.typhi.rep1

## ----extractData1, warning=FALSE----------------------------------------------
baseCalled(s.typhi.rep1)

## ----extractData2, warning=FALSE----------------------------------------------
baseCalled(s.typhi.rep1[1:2])

## ----summaryPlots1, fig.height=5, fig.width=12.5, dev='png', warning=FALSE----
p1 <- plotReadAccumulation(s.typhi.rep1)
p2 <- plotActiveChannels(s.typhi.rep1)
grid.arrange(p1, p2, ncol = 2)

## ----readCategories, fig.height=5, fig.width=12.5, dev='png', warning=FALSE----
p1 <- plotReadCategoryCounts(s.typhi.rep1)
p2 <- plotReadCategoryQuals(s.typhi.rep1)
grid.arrange(p1, p2, ncol = 2)

## ----summaryPlots2, fig.height=5, fig.width=12.5, dev='png', warning=FALSE----
p1 <- plotEventRate(s.typhi.rep1)
p2 <- plotBaseProductionRate(s.typhi.rep1)
grid.arrange(p1, p2, ncol = 2)

## ----numReads, fig.height=4.8, fig.width=14.5, dev='png', warning=FALSE-------
p1 <- layoutPlot(s.typhi.rep1, attribute = "nreads")
p2 <- layoutPlot(s.typhi.rep1, attribute = "kb")
grid.arrange(p1, p2, ncol = 2)

## ----heatmapExample, fig.height=4.8, fig.width=14.5, dev='png', message=FALSE, warning=FALSE----
library(dplyr)

read_count_2D <- baseCalled(s.typhi.rep1) %>% ## start with base called reads
  filter(strand == 'template') %>% ## keep template so we don't count things twice
  left_join(readInfo(s.typhi.rep1), by = 'id') %>% ## channel stored in @readInfo slot, match by id column
  group_by(channel) %>% ## group according to channel
  summarise(d2_count = length(which(full_2D == TRUE)), ## count those with full 2D status
            d2_prop = length(which(full_2D == TRUE)) / n()) ## divide by total count of reads from channel

## plot side-by-side
p1 <- channelHeatmap(read_count_2D, zValue = 'd2_count')
p2 <- channelHeatmap(read_count_2D, zValue = 'd2_prop')
grid.arrange(p1, p2, ncol = 2)

## ----channelActivity, fig.height=4.8, fig.width=10, dev='png', message=FALSE, warning=FALSE----
data(s.typhi.rep3, package = 'minionSummaryData')
## we will plot the median event signal for each read on z-axis
z_scale = select(eventData(s.typhi.rep3), id, median_signal)
channelActivityPlot( s.typhi.rep3, zScale = z_scale )

## ----pentamerCorrelation, fig.height=5, fig.width=6.5, dev='png', message=FALSE, warning=FALSE----
plotKmerFrequencyCorrelation( s.typhi.rep3, only2D = FALSE )

## ----combinedplots, eval=FALSE------------------------------------------------
# p1 <- channelActivityPlot(dat, select(eventData(dat), id, median_signal))
# p2 <- plotKmerFrequencyCorrelation(dat)
# grid.arrange(p1, p2, ncol = 2)

## ----writefastq, message=FALSE------------------------------------------------
library(ShortRead)
writeFastq( fastq( s.typhi.rep1 ), file = tempfile() )

## ----writefastq2--------------------------------------------------------------
writeFastq( fastq2D( s.typhi.rep1 ), file = tempfile() )

## ----fast5toFastq-------------------------------------------------------------
fast5files <- system.file('extdata', 'example.fast5', package = "IONiseR")
fast5toFastq(files = fast5files, fileName = "test", outputDir = tempdir(), 
             strand = 'all', ncores = 1)
list.files(path = tempdir(), pattern = "*.fq.gz$")

#' @import bit64
.getRawSummary <- function(file, readNumber = NA, dontCheck = FALSE) {
    
    if(is.character(file)) {
        fid <- H5Fopen(file)
        on.exit(H5Fclose(fid))
    } else {
        fid <- file
    }
    
    if(dontCheck || .groupExistsObj(fid, group = "/Raw/Reads/")) {
        ## get the Read_No. if we need to
        if(is.na(readNumber)) {
            readNumber <- .getReadNumber(fid)
        }
        
        ## Open the group and read the two attributes we want
        #gid <- H5Gopen(fid, paste0("/Raw/Reads/Read_", readNumber)) 
        #aid <- H5Aopen(gid, "duration")
        #duration <- H5Aread(aid) 
        #H5Aclose(aid)
        #aid <- H5Aopen(gid, "start_time") 
        #start_time <- H5Aread(aid) 
        #H5Aclose(aid)   
        
        did <- H5Dopen(fid, paste0("/Raw/Reads/Read_", readNumber, "/Signal"))
        #signal <- H5Dread(did, bit64conversion = "int", compoundAsDataFrame = FALSE)
        signal <- H5Dread(did, compoundAsDataFrame = FALSE)
        H5Dclose(did)
        mean_signal <- mean(signal)
        #duration <- length(signal)

        #H5Gclose(gid)
    } else {
        mean_signal <- NA
    } 
    
    return(tibble(mean_signal))  
}

.getEventsSummary <- function(file, readNumber = NA, dontCheck = FALSE) {
    
    if(is.character(file)) {
        fid <- H5Fopen(file)
        on.exit(H5Fclose(fid))
    } else {
        fid <- file
    }

    if(dontCheck || .groupExistsObj(fid, group = "/Analyses/EventDetection_000/Reads")) {
        ## get the Read_No. if we need to
        if(is.na(readNumber)) {
            readNumber <- .getReadNumber(fid)
        }

        ## Open the group and read the attributes we want
        gid <- H5Gopen(fid, paste0("/Analyses/EventDetection_000/Reads/Read_", readNumber)) 
        start_time <- tryCatch(.readAttribute(gid, attribute = "start_time"), 
                             error = function(e) { NA })
        duration <- tryCatch(.readAttribute(gid, attribute = "duration"), 
                             error = function(e) { NA })
        H5Gclose(gid)
        
        if(.groupExistsObj(fid, paste0("/Analyses/EventDetection_000/Summary/event_detection"))) {
            gid <- H5Gopen(fid, paste0("/Analyses/EventDetection_000/Summary/event_detection")) 
            num_events <- tryCatch(.readAttribute(gid, attribute = "num_events"), 
                                 error = function(e) { NA }) 
            H5Gclose(gid)
        } else {
            did <- H5Dopen(fid, paste0("/Analyses/EventDetection_000/Reads/Read_", readNumber, "/Events"))
            h5dataset <- H5Dget_space(did)    
            num_events <- H5Sget_simple_extent_dims(h5dataset)$size
            H5Dclose(did)
        }
    } else {
        start_time <- duration <- num_events <- NA
    }
    
    return(tibble(start_time, duration, num_events))  
}

.readAttribute <- function(gid, attribute) {
    aid <- H5Aopen(gid, attribute)
    on.exit( H5Aclose(aid) )
    val <- H5Aread(aid)
    return(val)
}


.getBaseCalledSummary <- function(file, strand = "template", d = "1D", analysisNum = "000",
                                  dontCheck = FALSE) {
    
    if(is.character(file)) {
        fid <- H5Fopen(file)
        on.exit(H5Fclose(fid))
    } else {
        fid <- file
    }
    
    if(dontCheck || .groupExistsObj(fid, group = paste0("/Analyses/Basecall_", d, "_", analysisNum, "/Summary/basecall_1d_", strand))) {
        ## Open the group and read the attribute we want
        gid <- H5Gopen(fid, paste0("/Analyses/Basecall_", d, "_", analysisNum, "/Summary/basecall_1d_", strand))
        
        num_events <- tryCatch(.readAttribute(gid, attribute = "num_events"), 
                               error = function(e) { NA })
        num_skips <- tryCatch(.readAttribute(gid, attribute = "num_skips"), 
                              error = function(e) { NA })
        num_stays <- tryCatch(.readAttribute(gid, attribute = "num_stays"), 
                              error = function(e) { NA })
        
        H5Gclose(gid)
    } else {
        num_events <- num_skips <- num_stays <- NA
    }
    basecalledStats <- tibble(num_events, num_skips, num_stays, strand)
    
    return(basecalledStats)  
}

#' IONiseR: A package for assessing quality of MinION data
#' 
#' IONiseR provides tools for the quality assessment of Oxford Nanopore MinION 
#' data. It extracts summary statistics from a set of fast5 files and can be 
#' used either before or after base calling.  In addition to standard summaries
#' of the read-types produced, it provides a number of plots for visualising 
#' metrics relative to experiment run time or spatially over the surface of a 
#' flowcell.
#'
#' @docType package
#' @name IONiseR
#' 
#' @import ggplot2
#' @import rhdf5
#' 
#' @importFrom methods setClass setGeneric setMethod
#' @importFrom magrittr "%>%"
#' @importFrom ShortRead strand
#' 
#' @importClassesFrom ShortRead ShortReadQ
NULL

#' @describeIn Fast5Summary Subset object and return an object of the same class.
#' 
#' @param x Object of class Fast5Summary
#' @param i Vector defining index to subset by.
#' 
#' @docType methods
#' @export
#' 
#' @importFrom dplyr select slice
#' @importFrom tidyr gather
#' @importFrom XVector compact
#' @importFrom methods initialize .hasSlot
setMethod("[", c("Fast5Summary", "ANY"), function(x, i) {
    recordTable <- .matchRecords(x)[i,]
    
    ## there are several columns holding information on baseCalled and fastq
    ## we do some manipulation here to get a single index list.
    baseCalledIDX <- select(recordTable, baseCalledTemplate:baseCalledComplement) %>% 
        gather(key = component, value = idx) %>% 
        filter(!is.na(idx))
    baseCalledIDX <- baseCalledIDX[['idx']]
    
    fastqIDX <- select(recordTable, fastqTemplate:fastq2D) %>% 
        gather(key = component, value = idx) %>% 
        filter(!is.na(idx))
    fastqIDX <- fastqIDX[['idx']]
    
    if(.hasSlot(x, 'versions')) {
        versions <- x@versions
    } else {
        versions <- list()
    }
    
    initialize(x,
               readInfo = slice(readInfo(x), recordTable[['readInfo']]),
               eventData = slice(eventData(x), recordTable[['eventData']]),
               baseCalled = slice(baseCalled(x), baseCalledIDX),
               fastq = XVector::compact(fastq(x)[fastqIDX]),
               versions = versions)
})


#' @importFrom ShortRead id
.idFromFASTQ <- function(fastq_obj) {
    as.integer(do.call(rbind, strsplit(as.character(ShortRead::id(fastq_obj)), "_"))[,1])
}

#' @importFrom ShortRead id
.readtypeFromFASTQ <- function(fastq_obj) {
    do.call(rbind, strsplit(as.character(ShortRead::id(fastq_obj)), "_"))[,2]
}

.matchRecords <- function(summaryData) {
    
    ids <- readInfo(summaryData)[['id']]
    ri_row <- match(ids, readInfo(summaryData)[['id']])
    ed_row <- match(ids, eventData(summaryData)[['id']])
    bct_row <- match(ids, filter(baseCalled(summaryData), strand == "template")[['id']])
    bcc_row <- match(ids, filter(baseCalled(summaryData), strand == "complement")[['id']]) + 
        nrow(filter(baseCalled(summaryData), strand == "template"))
    
    fastq_id <- .idFromFASTQ( fastq(summaryData) )
    fastq_id <- fastq_id[-(1:nrow(summaryData@baseCalled))]
    fq_row <- match(ids, fastq_id) + nrow(summaryData@baseCalled)
    
    record_table <- tibble(id = ids, 
                           readInfo = ri_row,
                           eventData = ed_row,
                           baseCalledTemplate = bct_row,
                           baseCalledComplement = bcc_row,
                           fastqTemplate = bct_row,
                           fastqComplement = bcc_row,
                           fastq2D = fq_row)
    
    return(record_table)
    
}

.get2D <- function(summaryData) {
    
    ids <- filter(baseCalled(summaryData), strand == "template", full_2D == TRUE)[,id]
    idx <- which(readInfo(summaryData)[,id] %in% ids)
    return(summaryData[idx,])
}


#' Extract template reads
#' 
#' This generic function accesses the fastq slot stored in an object derived 
#' from the Fast5Summary class, and returns only the template reads.
#' 
#' @param x Object of class \code{\linkS4class{Fast5Summary}}
#' @return A ShortReadQ object
#' @examples
#' if( require(minionSummaryData) ) {
#'    data(s.typhi.rep2, package = 'minionSummaryData')
#'    fastqTemplate( s.typhi.rep2 )
#' }
setGeneric("fastqTemplate", function(x) {
    standardGeneric("fastqTemplate")
})

#' Extract complement reads
#' 
#' This generic function accesses the fastq slot stored in an object derived 
#' from the Fast5Summary class, and returns only the complement reads.
#' 
#' @param x Object of class \code{\linkS4class{Fast5Summary}}
#' @return A ShortReadQ object
#' @examples
#' if( require(minionSummaryData) ) {
#'    data(s.typhi.rep2, package = 'minionSummaryData')
#'    fastqComplement( s.typhi.rep2 )
#' }
setGeneric("fastqComplement", function(x) {
    standardGeneric("fastqComplement")
})

#' Extract 2D reads
#' 
#' This generic function accesses the fastq slot stored in an object derived 
#' from the Fast5Summary class, and returns only the 2D reads.
#' 
#' @param x Object of class \code{\linkS4class{Fast5Summary}}
#' @return A ShortReadQ object
#' @examples
#' if( require(minionSummaryData) ) {
#'    data(s.typhi.rep2, package = 'minionSummaryData')
#'    fastq2D( s.typhi.rep2 )
#' }
setGeneric("fastq2D", function(x) {
    standardGeneric("fastq2D")
})

#' @describeIn Fast5Summary Returns ShortReadQ object containing only template reads
#' 
#' @export
setMethod("fastqTemplate", 
          c(x = "Fast5Summary"),
          function(x) {
              idx <- which(.readtypeFromFASTQ( fastq(x) ) == 'template')
              return( fastq(x)[idx,] )
          }
)

#' @describeIn Fast5Summary Returns ShortReadQ object containing only complement reads
#' 
#' @export
setMethod("fastqComplement", 
        c(x = "Fast5Summary"),
        function(x) {
            idx <- which(.readtypeFromFASTQ( fastq(x) ) == 'complement')
            return( fastq(x)[idx,] )
        }
)

#' @describeIn Fast5Summary Returns ShortReadQ object containing only 2D reads
#' 
#' @export
setMethod("fastq2D", 
          c(x = "Fast5Summary"),
          function(x) {
              idx <- which(.readtypeFromFASTQ( fastq(x) ) == '2D')
              return( fastq(x)[idx,] )
          }
)

#' Display correlation between pentemer proportions in two time windows
#' 
#' Plots 
#' 
#' @param summaryData Object of class \linkS4class{Fast5Summary}.
#' @param kmerLength Specifies the length of kmers to compare. Defaults to 5 
#' given the current pentamer reading nature of the nanopores.
#' @param groupedMinutes Defines how many minutes each grouping of reads spans.
#' @param only2D  Logical. If TRUE kmers are computed for only full 2D reads.
#' If FALSE 2D reads are ignored and all available template and complement 
#' strands are used.
#' @return Returns an object of class \code{gg} representing the plot.
#' @examples
#' if( require(minionSummaryData) ) {
#'    data(s.typhi.rep3, package = 'minionSummaryData')
#'    plotKmerFrequencyCorrelation( s.typhi.rep3, only2D = FALSE )
#' }
#' @export
#' @importFrom dplyr summarise_each funs arrange
#' @importFrom Biostrings oligonucleotideFrequency
#' @importFrom tidyr spread gather
#' @importFrom ShortRead sread
#' @importFrom stats cor
plotKmerFrequencyCorrelation <- function(summaryData, kmerLength = 5, groupedMinutes = 10, only2D = TRUE) {
    
    if(only2D) {
        idx <- grep('2D', id(fastq(summaryData)))
    } else {
        idx <- 1:nrow(baseCalled(summaryData))
    }
    pentamers <- oligonucleotideFrequency(x = sread(fastq(summaryData)[idx]), width = kmerLength, as.prob = TRUE)
    
    if(only2D) {
        tmp <- cbind(filter(baseCalled(summaryData), strand == "template", full_2D == TRUE), pentamers)
    } else {
        tmp <- cbind(baseCalled(summaryData), pentamers)
    }
    
    tmp2 <- group_by(tmp, time_group = start_time %/% (60 * groupedMinutes)) %>%
        summarise_each(funs(mean), AAAAA:TTTTT) %>%
        arrange(time_group) %>%
        gather(key = "pentamer", value = "freq", AAAAA:TTTTT) %>%
        spread(key = time_group, value = freq) %>% 
        select(-pentamer)
    
    correlations <- cor(as.matrix(tmp2))
    
    hours <- (as.integer(colnames(correlations)) * groupedMinutes) / 60
    tmp3 <- tibble(x = rep(hours, each = length(hours)), y = rep(hours, length(hours)), cor = as.vector(correlations))
    
    ggplot(tmp3, aes(x = x, y = y, fill = cor)) + 
        geom_raster() +
        scale_x_continuous(expand = c(0, 0)) + 
        scale_y_continuous(expand = c(0, 0)) +
        xlab("hour") +
        ylab("hour") +
        scale_fill_gradient(high = "orange", low = "darkblue")
}

## Check whether group is present in a fast5 file
## 
## file Character string containing the name of the file to be read.
## group Character string holding the group we are checking for.
.groupExistsString <- function(file, group) {
    fid <- H5Fopen(file)
    res <- H5Lexists(fid, group)
    H5Fclose(fid)
    return(res)
}

## Check whether group is present in a fast5 file
## 
## @param fid Object of class H5IdComponent created with H5Fopen().
## @param group Character string holding the group we are checking for.
.groupExistsObj <- function(fid, group) {
    res <- H5Lexists(fid, group)
    return(res)
}

## Some fast5 files can be corrupt.  This just checks whether we can open them
## and returns a boolean result.
#' @importFrom methods is
.checkOpening <- function(file) {
    fid <- .H5Fopen_tryCatch(file)
    res <- is(fid,"H5IdComponent")
    if(res) { H5Fclose(fid) }
    return(res)
}

.H5Fopen_tryCatch <- function(file) {
    out <- tryCatch(
        {
            H5Fopen(file)
        },
        error=function(cond) {
            message(paste('Error opening:', file, "- File Skipped"))
            return(NA)
        }
    )    
    return(out)
}

## try to guess version of fast5 file
#' @importFrom stringr str_detect
.fast5version <- function(file) {
    
    groups <- unique(h5ls(file)[,'group'])
    
    ## classic version with 2D slot
    if(any(str_detect("/Analyses/Basecall_2D_[0-9]{3}/BaseCalled_template", string = groups))) {
        version <- 1
        ## not often seen variant, seen near r9 chemistry release    
    } else if(any(str_detect("/Analyses/Basecall_RNN_1D_[0-9]{3}/BaseCalled_template", string = groups))) {
        version <- 3
        ## version produced since metrichor 1.16    
    } else if(any(str_detect("/Analyses/Basecall_1D_[0-9]{3}/BaseCalled_template", string = groups))) {
        version <- 2
        ## if none of the above are found, no basecalling has been performed    
    } else {
        version <- 0
    }
    
    return(version)
}

## Determines whether a read is in a pass or fail folder by looking at the path.
## If neither term is found NA is returned
.passFailStatus <- function(path) {
    if(!grepl('pass|fail', path)) {
        return(NA)
    } else {
        return( grepl('pass', path) )
    }
}

#' Extract readInfo slot
#' 
#' This generic function accesses the readInfo slot stored in an object 
#' derived from the Fast5Summary class.
#' 
#' @param x Object of class \code{\linkS4class{Fast5Summary}}
#' @return A data.frame with 5 columns
#' @examples
#' if( require(minionSummaryData) ) {
#'    data(s.typhi.rep2, package = 'minionSummaryData')
#'    readInfo( s.typhi.rep2 )
#' }
setGeneric("readInfo", function(x) {
    standardGeneric("readInfo")
})

#' @describeIn Fast5Summary Returns readInfo data.frame
#' 
#' @include classes.R
#' @export
setMethod("readInfo", 
          c(x = "Fast5Summary"),
          function(x) {
              x@readInfo
          }
)


#' Extract eventData slot
#' 
#' This generic function accesses the eventData slot stored in an object derived 
#' from the Fast5Summary class.
#' 
#' @param x Object of class \code{\linkS4class{Fast5Summary}}
#' @return A data.frame with 5 columns
#' @examples
#' if( require(minionSummaryData) ) {
#'    data(s.typhi.rep2, package = 'minionSummaryData')
#'    eventData( s.typhi.rep2 )
#' }
setGeneric("eventData", function(x) {
    standardGeneric("eventData")
})

#' @describeIn Fast5Summary Returns eventData data.frame
#' 
#' @include classes.R
#' @export
setMethod("eventData", 
          c(x = "Fast5Summary"),
          function(x) {
              x@eventData
          }
)

#' Extract baseCalled slot
#' 
#' This generic function accesses the baseCalled slot stored in an object 
#' derived from the Fast5Summary class.
#' 
#' @param x Object of class \code{\linkS4class{Fast5Summary}}
#' @return A data.frame with 6 columns
#' @examples
#' if( require(minionSummaryData) ) {
#'    data(s.typhi.rep2, package = 'minionSummaryData')
#'    baseCalled( s.typhi.rep2 )
#' }
setGeneric("baseCalled", function(x) {
    standardGeneric("baseCalled")
})


#' @describeIn Fast5Summary Returns baseCalled data.frame
#' 
#' @include classes.R
#' @export
setMethod("baseCalled", 
          c(x = "Fast5Summary"),
          function(x) {
              x@baseCalled
          }
)

#' Extract fastq slot
#' 
#' This generic function accesses the fastq slot stored in an object 
#' derived from the Fast5Summary class.
#' 
#' @param x Object of class \code{\linkS4class{Fast5Summary}}
#' @return A ShortReadQ object
#' @examples
#' if( require(minionSummaryData) ) {
#'    data(s.typhi.rep2, package = 'minionSummaryData')
#'    fastq( s.typhi.rep2 )
#' }
setGeneric("fastq", function(x) {
    standardGeneric("fastq")
})

#' @describeIn Fast5Summary Returns ShortReadQ object stored in fastq slot.
#' 
#' @include classes.R
#' @export
setMethod("fastq", 
          c(x = "Fast5Summary"),
          function(x) {
              x@fastq
          }
)

#' @importFrom stringr str_detect
#' @importFrom dplyr transmute slice select n
#' @noRd 
.strandExistence <- function(ls, strand = "BaseCalled_template") {
    
    loc <- filter(ls, name == strand) %>% 
            select(group, name) %>% 
            transmute(path = paste(group, name, sep = "/")) %>% 
            slice(n())
    loc <- ifelse(stringr::str_detect(loc, strand), loc[[1]], "")
    return(loc)
}

.chooseStrand <- function(paths) {
    matches <- str_match(string = paths, pattern = "(^.*)([12]D_)([0-9]+)(.*$)")
    matches <- matches[which(!is.na(matches[,1])),,drop = FALSE]

    if(!nrow(matches)) {
        return("")
    }
    
    ## if the we have 1D and 2D analysis workflows, throw an error.  
    ## These are really different!
    if(length(unique(matches[,3])) != 1) {
        stop("Inconsistent analysis workflows detected.  ",
             "Were these files analyses with different versions of MinKNOW?")
    }
    
    if(length(unique(matches[,4])) != 1) {
        ## if there's a mix of anaylsis numbers, warn we're picking the lowest
        warning("Inconsistent number of analysis runs detected between files.",
                " Defaulting to the earliest", call. = FALSE)
    } 
    return( paste0(matches[1,2], matches[1,3], 
                   sort(unique(matches[,4]))[1], matches[1,5]) )
}

#' Determines the 'processing status' of the files that have been supplied.
#' 
#' Fast5 files can be obtained at several points in the standard processing
#' flow.  Exactly which point was reached determines the type of data that is
#' present in the files e.g. raw signal, events detected, bases called etc.
#' This function looks at the hdf5 structure of the files to try and determine
#' what to expect.  Knowing about the structure can improve the performance of
#' data extraction code, since we don't necessarily have to check for 
#' existance every time.  This also trys to determine whether the '1D' or '2D'
#' workflow was run, since this also influences the path to (and existance of)
#' the template (and complement) data.#' 
#' 
#' @importFrom stringr str_detect
#' @keywords internal
#' @noRd 
.fast5status <- function(files, warn = FALSE) {
    
    ## is the read number present in the file name?
    readInName <- stringr::str_detect(string = files, pattern = "read[_]?([0-9]+)")
    
    lsList <- lapply(files, h5ls, recursive = 3, datasetinfo = FALSE)
    
    ## is /Raw/Reads present?
    rawReads <- sapply(lsList, FUN = function(x) {
        select(x, group) %>% 
            str_detect(pattern = "/Raw/Reads")
    })
    
    ## is /Analysis/EventDetection_000 present?
    eventDetection <- sapply(lsList, FUN = function(x) {
        select(x, group) %>% 
            str_detect(pattern = "/EventDetection")
    })
    
    ## is /Analysis/BaseCalled_2D present?
    basecalled_2d <- sapply(lsList, FUN = function(x) {
        select(x, name) %>% 
            str_detect(pattern = "BaseCalled_2D")
    })
    
    loc_template <- sapply(lsList, .strandExistence, strand = "BaseCalled_template")
    loc_template <- .chooseStrand(loc_template)
    
    loc_complement <- sapply(lsList, .strandExistence, strand = "BaseCalled_complement")
    loc_complement <- .chooseStrand(loc_complement)
    
    loc_2D <- sapply(lsList, .strandExistence, strand = "BaseCalled_2D")
    loc_2D <- .chooseStrand(loc_2D)

    return(list(read_in_name = all(readInName),
                raw_reads = all(rawReads),
                event_detection = all(eventDetection),
                loc_template = loc_template,
                loc_complement = loc_complement,
                loc_2D = loc_2D,
                basecalled_2d = any(basecalled_2d))) 
    
}

#' Read summary data from fast5 files.
#' 
#' Reads one or more fast5 files and collects summary information about them.  
#' 
#' Currently this function assumes all files passed to it come from the same 
#' sequencing run.  It makes no effort to check for alternative file names or 
#' the like.  If files from multiple runs are passed to it they will be 
#' collated together and any analysis performed on them will represent the 
#' mixture of both experiments.
#' 
#' @param files Character vector of fast5 files to be read.
#' @return Object of class \linkS4class{Fast5Summary}
#' @examples \dontrun{
#' fast5files <- list.files('/foo/bar/', pattern = '.fast5$')
#' summaryData <- readFast5Summary(fast5files)
#' }
#' @export
#' @importFrom tibble tibble as_tibble
#' @importFrom dplyr filter
#' @importFrom ShortRead append
#' @importFrom utils packageVersion
readFast5Summary <- function(files) {
    
    ## some files can't be opened, so we filter them here
    message("Checking file validity")
    files <- files[file.exists(files)]
    if(length(files) == 0) {
        stop('None of the provided files can be accessed.  ',
             'Have you supplied the correct path?')
    }
    fileStatus <- sapply(files, .checkOpening, USE.NAMES = FALSE)
    files <- files[ which(fileStatus) ]
    
    status <- .fast5status(files = sample(files, size = min(length(files), 15)))
    if(status$loc_template == "") {
        stop("No basecalls for template strand found.  Aborting")
    }
    
    message("Reading Channel Data")
    if(status$read_in_name) {
        readNums <- as.integer(str_match(string = files, pattern = "read[_]?([0-9]+)")[,2])
    } else { 
        readNums <- rep(NA, length(files))
    }
    readInfo <- do.call("rbind", mapply(.getReadChannelMux2, files, readNums, dontCheck = FALSE,
                                        USE.NAMES = FALSE, SIMPLIFY = FALSE))
    readInfo <- as_tibble(cbind(id = 1:nrow(readInfo), file = basename(files), readInfo))
    
    if(status$raw_reads) {
        message("Reading Raw Data")
        ## raw data is the median signal, and the duration of the read
        rawData <- as_tibble(do.call("rbind", mapply(.getRawSummary, files, readNums, dontCheck = TRUE, 
                                                     USE.NAMES = FALSE, SIMPLIFY = FALSE)))
    } 
    
    if(status$event_detection) {
        message("Reading event data")
        eventData <- do.call("rbind", mapply(.getEventsSummary, files, readNums, dontCheck = TRUE, 
                                             USE.NAMES = FALSE, SIMPLIFY = FALSE))
    } else {
        message("Event data not found")
        eventData <- NA
        eventData <- do.call("rbind", mapply(.getRawStartDuration, files, readNums, 
                                             USE.NAMES = FALSE, SIMPLIFY = FALSE))
        colnames(eventData) <- c("start_time", "duration")
    }
    
    if(status$raw_reads) {
        rawEventData <- as_tibble(cbind(id = readInfo[['id']], rawData, eventData))
    } else {
        rawEventData <- as_tibble(cbind(id = readInfo[['id']], eventData))
    }
    
    ## we convert timing data into seconds. 
    ## To do this we find the sampling rate stored in one file
    ## not possible if the event_data wasn't present in the files
    #if(status$event_detection) {
        samplingRate <- .getSamplingRate(files[1])
        rawEventData <- mutate(rawEventData, 
                               start_time = start_time / samplingRate,
                               duration = duration / samplingRate)
    #}
    
    message("Reading Template Data")
    d <- str_match(pattern = "_([12]D)_", string = status$loc_template)[,2]
    template <- do.call("rbind", mapply(.getBaseCalledSummary, files, dontCheck = FALSE, 
                                        strand = "template", d = d,
                                        SIMPLIFY = FALSE, USE.NAMES = FALSE))
    template <- mutate(template, id = readInfo[['id']])
    template <- filter(template, !(is.na(num_events)))
    
    message("Reading Template FASTQ")
    ## get the fastq for those that have it
    fq_t <- mapply(.getFastqString, files[ template[['id']] ], strand = "template", d = d)
    fq_t <- .processFastqVec(fq_t, readIDs = template[['id']], appendID = "_template")  
    fastq_template <- fq_t$fastq
    ## if there are any invalid entries we need to remove them
    if(length(fq_t$invalid)) {
        template <- template[-fq_t$invalid,]
    }
    baseCalled <- template
    
    if(nchar(status$loc_complement)) { 
        message("Reading Complement Data")
        d <- str_match(pattern = "_([12]D)_", string = status$loc_complement)[,2]
        complement <- do.call("rbind", mapply(.getBaseCalledSummary, files, dontCheck = FALSE, 
                                              strand = "complement", d = d,
                                              SIMPLIFY = FALSE, USE.NAMES = FALSE))
        complement <- mutate(complement, id = readInfo[['id']])
        complement <- filter(complement, !(is.na(num_events)))
        
        message("Reading Complement FASTQ")
        fq_c <- mapply(.getFastqString, files[ complement[['id']] ], strand = "complement", d = d)
        fq_c <- .processFastqVec(fq_c, readIDs = complement[['id']], appendID = "_complement")  
        fastq_complement <- fq_c$fastq
        
        ## if there are any invalid entries we need to remove them
        if(length(fq_c$invalid)) {
            complement <- complement[-fq_c$invalid,]
        }
        
    } else {
        fastq_complement <- ShortRead::ShortReadQ()
        complement <- NULL
    }
    
    ## we haven't read anything about 2D reads yet, so we need to identify which
    ## files have them.  Then we'll read only those
    idx2D <- which(sapply(files, .groupExistsString, group = paste0("/Analyses/Basecall_2D_000/BaseCalled_2D")))
    if(length(idx2D)) {
        message("Reading 2D FASTQ")
        fq_2D <- sapply(files[ idx2D ], .getFastqString, strand = "2D", d = "2D")
        fq_2D <- .processFastqVec(fq_2D, readIDs = readInfo[['id']][idx2D], appendID = "_2D")  
        fastq_2D <- fq_2D$fastq
    }
    
    ## We update the individual strands to indicate if they are part of a full 2D read
    template <- as_tibble(cbind(template, full_2D = template[['id']] %in% idx2D))
    complement <- as_tibble(cbind(complement, full_2D = complement[['id']] %in% idx2D))
    
    # ## combine the template, complement and 2D data
    baseCalled <- rbind(template, complement)
    fastq <- ShortRead::append(fastq_template, fastq_complement)
    if(length(idx2D)) {
        fastq <- ShortRead::append(fastq, fastq_2D)
    }
    
    message("Done")
    obj <- new("Fast5Summary", 
               readInfo = readInfo, 
               rawData = tibble(), 
               eventData = rawEventData, 
               baseCalled = baseCalled, 
               fastq = fastq,
               versions = list('IONiseR' = strsplit(as.character(packageVersion("IONiseR")),".",fixed=TRUE)[[1]]
                               # 'MinKNOW' = max(versions))
               )
    )
    
    return(obj)
}

#' @importFrom BiocParallel MulticoreParam register bpmapply
readFast5Summary.mc <- function(files, ncores = 2) {
  
  register(MulticoreParam(workers = ncores))
  
  ## some files can't be opened, so we filter them here
  message("Checking file validity")
  fileStatus <- sapply(files, .checkOpening, USE.NAMES = FALSE)
  files <- files[ which(fileStatus) ]
  
  status <- .fast5status(files = sample(files, size = min(length(files), 15)))
  if(status$loc_template == "") {
    stop("No basecalling for template strand found.  Aborting")
  }
  
  message("Reading Channel Data")
  if(status$read_in_name) {
    readNums <- as.integer(str_match(string = files, pattern = "read[_]?([0-9]+)")[,2])
  } else { 
    readNums <- rep(NA, length(files))
  }
  readInfo <- do.call("rbind", bpmapply(.getReadChannelMux2, files, readNums, dontCheck = TRUE,
                                      USE.NAMES = FALSE, SIMPLIFY = FALSE))
  readInfo <- as_tibble(cbind(id = 1:nrow(readInfo), file = basename(files), readInfo))
  
  if(status$raw_reads) {
    message("Reading Raw Data")
    ## raw data is the median signal, and the duration of the read
    rawData <- as_tibble(do.call("rbind", bpmapply(.getRawSummary, files, readNums, dontCheck = TRUE, 
                                                 USE.NAMES = FALSE, SIMPLIFY = FALSE)))
  } 
  
  message("Reading Event Data")
  eventData <- do.call("rbind", bpmapply(.getEventsSummary,  files, readNums, dontCheck = TRUE, 
                                       USE.NAMES = FALSE, SIMPLIFY = FALSE))
  
  if(status$raw_reads) {
    rawEventData <- as_tibble(cbind(id = readInfo[['id']], rawData, eventData))
  } else {
    rawEventData <- as_tibble(cbind(id = readInfo[['id']], eventData))
  }
  
  ## we convert timing data into seconds. 
  ## To do this we find the sampling rate stored in one file
  samplingRate <- .getSamplingRate(files[1])
  rawEventData <- mutate(rawEventData, 
                         start_time = start_time / samplingRate,
                         duration = duration / samplingRate)
  
  message("Reading Template Data")
  d <- str_match(pattern = "_([12]D)_", string = status$loc_template)[,2]
  template <- do.call("rbind", bpmapply(.getBaseCalledSummary, files, dontCheck = FALSE, 
                                      strand = "template", d = d,
                                      SIMPLIFY = FALSE, USE.NAMES = FALSE))
  template <- mutate(template, id = readInfo[['id']])
  template <- filter(template, !(is.na(num_events)))
  
  message("Reading Template FASTQ")
  ## get the fastq for those that have it
  fq_t <- bpmapply(.getFastqString, files[ template[['id']] ], strand = "template", d = d)
  fq_t <- .processFastqVec(fq_t, readIDs = template[['id']], appendID = "_template")  
  fastq_template <- fq_t$fastq
  ## if there are any invalid entries we need to remove them
  if(length(fq_t$invalid)) {
    template <- template[-fq_t$invalid,]
  }
  baseCalled <- template
  
  if(nchar(status$loc_complement)) { 
    message("Reading Complement Data")
    d <- str_match(pattern = "_([12]D)_", string = status$loc_complement)[,2]
    complement <- do.call("rbind", bpmapply(.getBaseCalledSummary, files, dontCheck = FALSE, 
                                          strand = "complement", d = d,
                                          SIMPLIFY = FALSE, USE.NAMES = FALSE))
    complement <- mutate(complement, id = readInfo[['id']])
    complement <- filter(complement, !(is.na(num_events)))
    
    
    message("Reading Complement FASTQ")
    fq_c <- bpmapply(.getFastqString, files[ complement[['id']] ], strand = "complement", d = d)
    fq_c <- .processFastqVec(fq_c, readIDs = complement[['id']], appendID = "_complement")  
    fastq_complement <- fq_c$fastq
    
    ## if there are any invalid entries we need to remove them
    if(length(fq_c$invalid)) {
      complement <- complement[-fq_c$invalid,]
    }
    
  } else {
    fastq_complement <- ShortRead::ShortReadQ()
    complement <- NULL
  }
  
  ## we haven't read anything about 2D reads yet, so we need to identify which
  ## files have them.  Then we'll read only those
  idx2D <- which(sapply(files, .groupExistsString, group = paste0("/Analyses/Basecall_2D_000/BaseCalled_2D")))
  if(length(idx2D)) {
    message("Reading 2D FASTQ")
    fq_2D <- sapply(files[ idx2D ], .getFastqString, strand = "2D", d = "2D")
    fq_2D <- .processFastqVec(fq_2D, readIDs = readInfo[['id']][idx2D], appendID = "_2D")  
    fastq_2D <- fq_2D$fastq
  }
  
  ## We update the individual strands to indicate if they are part of a full 2D read
  template <- as_tibble(cbind(template, full_2D = template[['id']] %in% idx2D))
  complement <- as_tibble(cbind(complement, full_2D = complement[['id']] %in% idx2D))
  
  # ## combine the template, complement and 2D data
  baseCalled <- rbind(template, complement)
  fastq <- ShortRead::append(fastq_template, fastq_complement)
  if(length(idx2D)) {
    fastq <- ShortRead::append(fastq, fastq_2D)
  }
  
  message("Done")
  obj <- new("Fast5Summary", 
             readInfo = readInfo, 
             rawData = tibble(), 
             eventData = rawEventData, 
             baseCalled = baseCalled, 
             fastq = fastq,
             versions = list('IONiseR' = strsplit(as.character(packageVersion("IONiseR")),
                                                  ".",
                                                  fixed=TRUE)[[1]]
                             # 'MinKNOW' = max(versions))
             )
  )
  
  return(obj)
}

library(IONiseR)
context('Testing methods for the Fast5Summary class')

require(minionSummaryData)
data(s.typhi.rep2, package = "minionSummaryData")

test_that("length is the number of files", {
  expect_equal(length(s.typhi.rep2), 3591)
})

test_that("dim is not defined", {
  expect_equal(dim(s.typhi.rep2), NULL)
})

test_that("subsetting works", {
    expect_equal(length(s.typhi.rep2[1:10]), 10)
})

fast5files <- system.file('extdata', c('example.fast5', 'example_v2.fast5'), package = "IONiseR")

test_that("Error thrown when incompatible files combined", {
    expect_error(readFast5Summary(fast5files), "Inconsistent analysis workflows")
})

test_that("Error thrown when file can't be found", {
    expect_error(readFast5Summary('/not/path/to/file.fast5'), "None of the provided files can be accessed")
})

f1 <- readFast5Summary(fast5files[2])

test_that("Can fast5 file be read?", {
  expect_is(f1, 'Fast5Summary')
})

test_that("accessors", {
    expect_is(readInfo(f1), 'data.frame')
    expect_is(eventData(f1), 'data.frame')
    expect_true(all(dim(eventData(f1)) == c(1,4)))
    expect_is(baseCalled(f1), 'data.frame')
    expect_is(fastq(f1), 'ShortReadQ')
})

test_that("Show method prints summary", {
    expect_output(show(f1), regexp = "^Object of class: Fast5Summary")
})

test_that("Reading FASTQ", {
    expect_equal(length(fastq2D(f1)), 1)
    expect_equal(length(fastqTemplate(f1)), 1)
    expect_equal(length(fastqComplement(f1)), 1)
})

test_that("Catch broken file", {
    fast5_nw <- system.file('extdata', 'example_not-working.fast5', package = "IONiseR")
    expect_error(readFast5Summary(fast5_nw), "No basecalls for template strand found")
})

# library(rhdf5)
# library(dplyr)
# library(tibble)
# library(ggplot2)
# library(lubridate)
# library(zoo)
# 
# .getSignalMetaData <- function(file) {
#     
#     dat <- h5readAttributes(file, "/UniqueGlobalKey/channel_id/")
#     dat$sampling_rate <- IONiseR:::.getSamplingRate(file)
#     dat$exp_start_time <- as.numeric(h5readAttributes(file, "/UniqueGlobalKey/tracking_id")$exp_start_time)
#     return( dat )
# }
# 
# .getEvents <- function(file) {
#     
#     fid <- H5Fopen(file)
#     on.exit(H5Fclose(fid))
#     
#     exists <- IONiseR:::.groupExistsObj(fid, group = "/Analyses/EventDetection_000/Reads")
#     if(!exists) {
#         start_time <- duration <- num_events <- median_signal <- NA
#     } else {
#         ## get the Read_No., this changes in every file
#         gid <- H5Gopen(fid, "/Analyses/EventDetection_000/Reads")
#         read_number_char <- h5ls(gid)[1,"name"]
#         H5Gclose(gid)
#         
#         ## Open the group
#         gid <- H5Gopen(fid, paste0("/Analyses/EventDetection_000/Reads/", read_number_char)) 
#         did <- H5Dopen(gid, "Events")
#         events <- as_tibble(H5Dread(did, bit64conversion = "int", compoundAsDataFrame = TRUE))
#         H5Dclose(did)
#         
#         H5Gclose(gid)
#     } 
#     
#     return(events)
# }
# 
# .getRaw <- function(file) {
#     
#     fid <- H5Fopen(file)
#     on.exit(H5Fclose(fid))
#     
#     exists <- IONiseR:::.groupExistsObj(fid, group = "/Raw/Reads")
#     if(!exists) {
#         start_time <- duration <- num_events <- median_signal <- NA
#     } else {
#         ## get the Read_No., this changes in every file
#         gid <- H5Gopen(fid, "/Raw/Reads")
#         read_number_char <- h5ls(gid)[1,"name"]
#         H5Gclose(gid)
#         
#         ## Open the group
#         gid <- H5Gopen(fid, paste0("/Raw/Reads/", read_number_char)) 
#         did <- H5Dopen(gid, "Signal")
#         signal = H5Dread(did)
#         H5Dclose(did)
#         
#         ## get the starting time
#         aid <- H5Aopen(gid, name = "start_time")
#         startTime <- H5Aread(aid)
#         H5Aclose(aid)
#         H5Gclose(gid)
#         
#         raw <- tibble(signal, time = seq(startTime, length.out = length(signal)))
#     } 
#     
#     return(raw)
# }
# 
# # [experiment start time]	/UniqueGlobalKey/tracking_id/{exp_start_time}
# # [read start time]	/Raw/Reads/Read_<nnn>/{start_time}
# # [sampling rate]	/UniqueGlobalKey/channel_id/{sampling_rate}
# # [channels 0pA adc]	/UniqueGlobalKey/channel_id/{offset}
# # [digitisable range in pA]	/UniqueGlobalKey/channel_id/{range}
# # [digitisation]	/UniqueGlobalKey/channel_id/{digitisation}
# # x-axis: [signal time in UTC] = [experiment start time] + ([read start time] + <signal_index_0_based>) / [sampling rate]
# # y-axis: [current in pA] = (<signal_value> + [channels 0pA adc] ) * [digitisable range in pA] / [digitisation]
# # <signal_value> is the column value in the dataset at /Raw/Reads/Read_###/Signal
# # <signal_index_0_based> is the row index starting at 0 for the time conversion formula to work
# 
# f1 <- "~/projects/bioconductor/IONiseR/inst/extdata/example_v2.fast5"
# f1 <- "/mnt/data/randomEncounters/MichaelVeronesi/Nanopore Sample 2/graveley_HP_20160627_FN_MN16664_sequencing_run_sample_id_84084_ch276_read1110_strand.fast5"
# s1 <- .getSignalMetaData(f1)
# t1 <- .getRaw(f1) %>%
#     mutate(pA = signal + s1$offset * (s1$range / s1$digitisation),
#            utc = (time / s1$sampling_rate) + as.numeric(s1$exp_start_time))
# 
# slice(t1,500:5000 ) %>%
#     mutate(date = as_datetime(utc)) %>%
# ggplot( aes(x = date, y = pA)) + geom_line()
# 
# t2 <- slice(t1,2500:4000 ) %>%
#     mutate(date = as_datetime(utc)) %>%
#     mutate(mean.back = rollmean(x = pA, 15, align = "right", fill = NA),
#            mean.forward = rollmean(x = pA, 15, align = "left", fill = NA)) %>%
#     mutate(mad.back = 3 * rollapply(data = pA, width = 15, align = "right", fill = NA, FUN = mad)) %>%
#     mutate(mad.forward = 3 * rollapply(data = pA, width = 15, align = "left", fill = NA, FUN = mad)) %>%
#     mutate(jump = ifelse((pA<mean.back-mad.back|pA>mean.back+mad.back) &
#                         (pA>mean.forward-mad.forward&pA<mean.forward+mad.forward), 
#                         TRUE, FALSE)) %>%
#     mutate(jump2 = rollapply(data = jump, width = 15, align = "right", fill = FALSE, FUN = function(x) { ifelse(sum(x, na.rm=TRUE)==1&x[15], TRUE, FALSE)}))
# 
#     
# ggplot(t2, aes(x = date, y = pA)) + geom_point() +
#     geom_point(data = filter(t2, jump2==TRUE), aes(x = date, y = pA), color = "cyan")

#' An S4 class for summarised data from a MinION sequencing run
#'
#' @slot readInfo Object of class tibble.  Contains five columns: 
#' \itemize{
#'   \item id - an integer key that allows use to match entries in the separate 
#'   slots of this object.
#'   \item file - Basename of the fast5 file the data was read from.
#'   \item read - Read number from channel.
#'   \item channel - channel.
#'   \item mux - Specific pore that was used within the four that are assigned 
#'   to a single channel. Should be in the range 1-4, but if this isn't 
#'   available it will be 0.
#' }
#' @slot rawData Object of class tibble.  Intended to hold raw signal data
#' although reading this is currently not implemented in \code{IONiseR}.
#' @slot eventData Object of class tibble.  Holds summary of events data 
#' prior to base calling. Contains five columns: 
#' \itemize{
#'   \item id - an integer key that allows use to match entries in the 
#'   separate slots of this object.
#'   \item start_time - time in seconds after the run started that this 
#'   reading began. 
#'   \item duration - time in seconds the reading lasted. 
#'   \item num_events - the number of events that were recorded as part of 
#'   this reading. 
#'   \item median_signal - median of the recorded signals for this set of events.
#' }
#' @slot baseCalled Object of class tibble.  For the most part contains 
#' similar data to the @@eventData slot, the base called data is derived from it.
#' \itemize{
#'   \item id - an integer key that allows use to match entries in the 
#'   separate slots of this object.
#'   \item start_time - time in seconds after the run started that this 
#'   reading began. 
#'   \item duration - time in seconds the reading lasted. 
#'   \item num_events - the number of events that were recorded as part of 
#'   this reading. 
#'   \item strand - can be either 'template' or 'complement'
#'   \item full_2D - boolean value specifying whether the read forms part of a 
#'   2D pair.  If TRUE the FASTQ data for the template, complement and 2D read 
#'   will be available in the @@fastq slot.
#' }
#' @slot fastq Object of class ShortReadQ.  This slot contains all reads 
#' (template, complement and 2D).  The read names take the form NUM_STRAND, 
#' where NUM matches with the id column in the other slots and STRAND indicates 
#' whether the read is template, complement or 2D.
#' @slot versions A list intended to store the version of IONiseR that was used
#' to create the object.  (May be extended in the future to include the version
#' of MinKNOW that the original fast5 files were processed, if this can be 
#' determined accurately.)
#' @return An object of class Fast5Summary
#' @name Fast5Summary-class
#' @exportClass Fast5Summary
setClass("Fast5Summary",
         slots = list(readInfo = "tbl_df",
                      rawData = "tbl_df",
                      eventData = "tbl_df",
                      baseCalled = "tbl_df",
                      fastq = "ShortReadQ",
                      versions = "list"))



setMethod(show, "Fast5Summary", function(object) {
    cat("Object of class: Fast5Summary\nContains information from:\n")
    cat(" ", nrow(readInfo(object)), "fast5 files\n")
    cat("  |-", nrow(filter(baseCalled(object), strand == "template")), "template strands\n")
    cat("  |-", nrow(filter(baseCalled(object), strand == "complement")), "complement strands\n")
    cat("  |-", nrow(filter(baseCalled(object), strand == "template", 
                            full_2D == TRUE)), "full 2D reads\n")
    ## if we have pass/fail information, print that too
    ## we also check that the pass column exists, hopefully we can remove this at some point
    if("pass" %in% names(readInfo(object))) {
        pf <- nrow(filter(readInfo(object), !is.na(pass)))
        if(pf) {
            cat("  |-", nrow(filter(readInfo(object), pass==TRUE)), "pass reads")
        }
    }
})

#' @describeIn Fast5Summary Returns the number of files read during creation of the object
#' @examples
#' if( require(minionSummaryData) ) {
#'    data(s.typhi.rep2, package = 'minionSummaryData')
#'    length( s.typhi.rep2 )
#' }
#' @export
setMethod(length, "Fast5Summary", function(x) {
    nrow( readInfo(x) )
})

---
title: "IONiseR"
author: "Mike L. Smith"
date: "`r doc_date()`"
package: "`r pkg_ver('IONiseR')`"
output: 
  BiocStyle::html_document
vignette: >
  %\VignetteIndexEntry{Quality assessment tools for nanopore data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}  
bibliography: bibliography.bib
---

```{r style, echo = FALSE, results = 'asis'}
BiocStyle::markdown()
knitr::opts_chunk$set(eval=TRUE)
```

# Introduction

This package is intended to provide tools for the quality assessment of data produced by Oxford Nanopore's MinION sequencer.  It includes a functions to generate a number plots for examining the statistics that we think will be useful for this task.

However, nanopore sequencing is an emerging and rapidly developing technology.  It is not clear what will be most informative.  We hope that `IONiseR` will provide a framework for visualisation of metrics that we haven't thought of, and welcome feedback at mike.smith@embl.de.

If you're not interested in the quality assement of the raw or event level data, and want to jump straight to the getting FASTQ format files from fast5 files you can go straight to the final section of this document.

## Getting started

In order to get started we need to load the `IONiseR` library.  In addition also load `ggplot2` and `gridExtra`, which are useful for arranging the plots in this vignette, but they are not essential to using the package.

```{r loadLibs, message=FALSE, warning=FALSE}
library(IONiseR)
library(ggplot2)
library(gridExtra)
```

# Reading data

Once the libraries are loaded we need to read some data.  We do this using the function `readFast5Summary()`.  This function takes a vector containing the path names of fast5 files you'd like to read.  The example below looks in a specific folder and selects all of the files whose name ends with ".fast5".  We then pass this list of files to the reading function.

*You should replace "/path/to/data" with the location of your fast5 files*.

```{r readData, eval=FALSE}
fast5files <- list.files(path = "/path/to/data/", pattern = ".fast5$", full.names = TRUE)
example.summary <- readFast5Summary( fast5files )
```

Raw fast5 data isn't distributed with this package, but example of the summarised format can found in the accompanying `minionSummaryData` package.  The following command will load this data, giving us an object called `s.typhi.rep1`.  If you have your own MinION data you wish to work with you should ignore this section of code and modify the example above to read your own files.

*The data presented in this example are taken from the publication by Ashton* et al *[@ashton2015minion]. You can obtain the original data from the European Nucleotide Archive here: http://www.ebi.ac.uk/ena/data/view/ERR668747*

```{r loadData}
library(minionSummaryData)
data(s.typhi.rep1)
```

Typing the name of the resulting object will print a short summary of its contents to the screen.

```{r exampleData}
s.typhi.rep1
```

## The `Fast5Summary` class

The `s.typhi.rep1` object is an example of a the `Fast5Summary` class.  

The structure of the class tries to reflect the variety of data one can find in fast5 files.  Depending upon the quality of the data basecalling may not be successful resulting in fast5 files that are essentially empty or files that contain event information but no called bases.  More commonly, basecalling is successful to some extent, but there is still a range of possibilities including files with only a template strand called, files with template and complement strands of differening lengths, and ideally files with well matched template and complement strands, plus a consensus 2D read.

*It is possible to read fast5 files that have not been basecalled, but since in most cases uploading to the Metrichor basecaller is performed automatically, it is unlikely that this will be a common usecase.*

The `Fast5Summary` class has four slots, each of which is designed to store data relating to one level of processing described above. The data themselves are stored as either as a `data.frame` or a `ShortReadQ`, one per slot. The table below gives the names of the four slots, along with a description of stage within the base calling process this represents and the specific fields that are currently stored in the appropriate `data.frame`. 

  Fast5Summary-class 
  -------------------------
  **readInfo** - All fast5 files contain this level of information 
  File name, channel, mux, pass/fail status
  **eventData** - If events were recorded this level is populated 
  Mean signal, start time, duration, no. of events 
  **baseCalled** - Created if base calling succeeded. Separate entries for each strand 
  Start time, duration, no. of events, strand, 2D status 
  **fastq** - Up to three entries per file (template, complement and 2D reads) 
  Bases and qualities 


### Extracting data

Data in the four slots can be obtained using the appropriate accessor methods: `readInfo()`, `eventData()`, `baseCalled()` and `fastq()`.  The example below extracts the base called from our example object.

```{r extractData1, warning=FALSE}
baseCalled(s.typhi.rep1)
```

### Subsetting

Since the the number of entries in each slot can be different between a specific reads, the `id` column is present in all entires corresponds to the fast5 file data was read from.  Subsetting operations work relative to this `id` field, so all data from the selected files is retained.  In the example below we select two files and can see that both the template and complement read information is retained in the `baseCalled` slot.

```{r extractData2, warning=FALSE}
baseCalled(s.typhi.rep1[1:2])
```

# Summary plots

## Data production

Once data have been read into a summary object `IONiseR` contains a number of functions for plotting the data.  The first example below visualises the accumulation of reads over the run time.  The second plot shows the how many channels were active (i.e. the number of molecules being read) during each minute of the experiment.

```{r summaryPlots1, fig.height=5, fig.width=12.5, dev='png', warning=FALSE}
p1 <- plotReadAccumulation(s.typhi.rep1)
p2 <- plotActiveChannels(s.typhi.rep1)
grid.arrange(p1, p2, ncol = 2)
```

## Read types

We may also be interested in the proportion of reads types that were generated.  Ideally, Oxford Nanopore's sequencing technology works by reading both the template and complement strands of a double-stranded DNA molecule.  The readings from both strands are then combined to give a higher confidence consensus sequence for the whole fragment - referred to as a 2D read.  

Given the nature of this process, there is a strict hierarchy to the data that can be found in a fast5 file. A full 2D read requires both a complement and template strand to have been read correctly. Similarly, a complement strand can only be present if the template was read successfully. Finally, you can encounter a file containing no called bases on either strand.

The function `plotReadCategories()` will visualise the total number of fast5 files summarised in an `Fast5Summary` object, along with the counts of those containing template, complement and 2D calls.  For an ideal dataset all four bars will be the same height, and the difference between them can reflect the quality of a dataset.  These values are the same as those printed out when typing in the name of a summary data object.

It may also be interesting to examine the base quality scores for the reads in the three categories.  The function `plotReadCategoryQuals()` allows one to do this, calculating the mean quality score for each sequence.

```{r readCategories, fig.height=5, fig.width=12.5, dev='png', warning=FALSE}
p1 <- plotReadCategoryCounts(s.typhi.rep1)
p2 <- plotReadCategoryQuals(s.typhi.rep1)
grid.arrange(p1, p2, ncol = 2)
```

## Reading rates

We can also look the performance of the pores over time.  For example, we may be interested in how rapidly events occur, which should be analagous to the rate at which molecules move through the nanopores.  The function `plotEventRate()` visualises this.

In similar fashion, we can also look at the rate at which actual nucleotide bases are called changes over time using `plotBaseProductionRate()`.  One would expect this to be closely related to the event rate (since each event should correspond to a base moving through the pore), however it is possible to envisage an scenario where events are recorded, but for some reason the base caller struggles to interpret them.

```{r summaryPlots2, fig.height=5, fig.width=12.5, dev='png', warning=FALSE}
p1 <- plotEventRate(s.typhi.rep1)
p2 <- plotBaseProductionRate(s.typhi.rep1)
grid.arrange(p1, p2, ncol = 2)
```

# Layout plots

When considering channel related metrics we can plot them as they are laid out on the flow cell.  To create plots of this form we can use the function `layoutPlot()`.  
```{r numReads, fig.height=4.8, fig.width=14.5, dev='png', warning=FALSE}
p1 <- layoutPlot(s.typhi.rep1, attribute = "nreads")
p2 <- layoutPlot(s.typhi.rep1, attribute = "kb")
grid.arrange(p1, p2, ncol = 2)
```

The `attribute` argument currently takes three possible values: "nreads", "kb" or "signal" which will respectively plot the total number of reads, the cumulative number of bases read by a channel, and the median signal recorded by the pore.  

## Plotting alternative metrics

If you wish to map a different metric on to the channel layout you can use the fuction `channelHeatmap()`.  This function requires a data.frame as input, with one column called 'channel'.  You can then use to the argument `zValue` to specify the intensity with which each channel is plotted.  

To demonstrate this, in the example below we will plot both the number of full 2D reads produced by each channel, and the proportion of all reads from the channel that are 2D.  When retrieving data from the `Fast5Summary`  object, each slot is returned as a `data.table`, allowing us to use `dplyr` and its associated pipe paradigm to manipulate the data. 

```{r heatmapExample, fig.height=4.8, fig.width=14.5, dev='png', message=FALSE, warning=FALSE}
library(dplyr)

read_count_2D <- baseCalled(s.typhi.rep1) %>% ## start with base called reads
  filter(strand == 'template') %>% ## keep template so we don't count things twice
  left_join(readInfo(s.typhi.rep1), by = 'id') %>% ## channel stored in @readInfo slot, match by id column
  group_by(channel) %>% ## group according to channel
  summarise(d2_count = length(which(full_2D == TRUE)), ## count those with full 2D status
            d2_prop = length(which(full_2D == TRUE)) / n()) ## divide by total count of reads from channel

## plot side-by-side
p1 <- channelHeatmap(read_count_2D, zValue = 'd2_count')
p2 <- channelHeatmap(read_count_2D, zValue = 'd2_prop')
grid.arrange(p1, p2, ncol = 2)
```

# Plotting against channel and time

If we want to look at the patterns that affect specific channels, or all channels for specific periods of time, we can use the function `channelActivityPlot()`.  By default this will plot a for every FAST5 file, sorted by channel on the y-axis and the time of the first and last recorded events on the x-axis.  

If provided with a `zScale` argument each line will be colored according to the specified data.  `zScale` expects to be passed a data.frame containing the id of a read, and a column containing the metric of interest.  In the example below we extract the id and median_signal columns from a Fast5Summary object.

```{r channelActivity, fig.height=4.8, fig.width=10, dev='png', message=FALSE, warning=FALSE}
data(s.typhi.rep3, package = 'minionSummaryData')
## we will plot the median event signal for each read on z-axis
z_scale = select(eventData(s.typhi.rep3), id, median_signal)
channelActivityPlot( s.typhi.rep3, zScale = z_scale )
```

Small numbers of reads with extreme values can compress the colours on the z-axis.  To make time related patterns easier to pick out reads across all channels are grouped by the starting time, and the mean value for all reads in the group is calculated and shown along the bottom of the plot.

# Pentamer content

Given that the sequence of a read is inferred from the recorded signal, one might wish to see if fluctuations in current over time are reflected in the base content of the reads produced.  The function `plotKmerFrequencyCorrelation()` breaks reads into groups by the time the read was first entered a pore.  The distribution of kmer (defaults to pentamers, but can be specified using the `kmerLength` argument) frequencies is then calculated for each window.  The correlation between each window and all others is then plotted, allowing one to see if the kmer content alters during run time.  The argument `only2D` switches between using the recorded template and complement strands, or the consensus 2D sequence.

```{r pentamerCorrelation, fig.height=5, fig.width=6.5, dev='png', message=FALSE, warning=FALSE}
plotKmerFrequencyCorrelation( s.typhi.rep3, only2D = FALSE )
```

The data included in `minionSummaryData` are very early in the life span of the MinION device, with a relatively smaller number of reads produced by only a few channels.  The following plots are created using a more recent, but currently unpublished, set of data.  Distinct waves in the average signal can be seen, although this doesn't seem to greatly impact the pentamer content of the resulting reads.

```{r combinedplots, eval=FALSE}
p1 <- channelActivityPlot(dat, select(eventData(dat), id, median_signal))
p2 <- plotKmerFrequencyCorrelation(dat)
grid.arrange(p1, p2, ncol = 2)
```

![Channel and Kmer correlation](ChannelKmer.png)

# Extracting reads

If you wish to write reads out to a FASTQ file there are two option you can use, depending upon whether you have used some of the previous quality assessment step, or simply want to go straight to the FASTQ files without looking at any of the other metrics.

## Writing from `Fast5Summary` object

If you already created a `Fast5Summary` object for your files, this contains the FASTQ entries and you can use the `writeFastq()` function from the `ShortRead` package to write them to disk, as shown here:

```{r writefastq, message=FALSE}
library(ShortRead)
writeFastq( fastq( s.typhi.rep1 ), file = tempfile() )
```

`IONiseR` also includes shortcut accessor functions for extracting the fastq entries for only the template (`fastqTemplate()`), complement (`fastqComplement()`) or 2D (`fastq2D()`) reads.  You can use these to extract and write out only the subset of reads you are interested in.  The example below will save only the 2D reads to a file.

```{r writefastq2}
writeFastq( fastq2D( s.typhi.rep1 ), file = tempfile() )
```

## Obtaining FASTQ directly from fast5

If you aren't interested in the quality control metrics, or have already assessed them, it is quicker to extract the FASTQ files directly and ignore the other data in the fast5 files.  You can do this using the function `fast5toFastq()`.  In the example shown here we work on a single fast5 file provided with the package, however you can provide a vector file names to the `files` argument and the FASTQ entries found in all of them will be combined together.

`fast5toFastq()` takes a number of additional arguments.  `fileName` provides a common stem for the created FASTQ files, while `outputDir()` defines the location the new files will be created in.  You can use `strand` to define which strands you want to extract and can take any combination of the options: *"template", "complement", "2D", "all"* and *"both"*.  The default is *"all"*, which will give you everything available. If you want only the template and complement strands you can use *both* or specify them explicitly.  `ncores` lets you specify the number of CPU cores to be used during the extraction.  This can potentially speed up this operation, but in my experience this seems to be more IO bound than CPU, so there is little benefit achieved by using a high number of cores.

The final command below lists the files that have been created by the `fast5toFastq()`  You can see that we have created three files, each of which start with the text given to the `fileName` argument and then the appropriate strand is appended to the name.

```{r fast5toFastq}
fast5files <- system.file('extdata', 'example.fast5', package = "IONiseR")
fast5toFastq(files = fast5files, fileName = "test", outputDir = tempdir(), 
             strand = 'all', ncores = 1)
list.files(path = tempdir(), pattern = "*.fq.gz$")
```

# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting_layout.R
\name{muxHeatmap}
\alias{muxHeatmap}
\title{Create layout plot of flowcell}
\usage{
muxHeatmap(data, zValue)
}
\arguments{
\item{data}{A data.frame.  Should have at least two columns, 
one of which has the name 'channel'.}

\item{zValue}{Character string specifying the name of the column to be used 
for the colour scaling.}
}
\value{
Returns an object of \code{gg} representing the plot.
}
\description{
Creates a plot representing the layout of a MinION flow cell.  Each circle
represents an individual channel with the intensity relecting a specified 
sequencing metric.  This function is a more generalised version of 
\code{\link{layoutPlot}}, allowing the user to map any value the like on 
the channel layout.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Methods-subsetting.R
\name{fastqTemplate}
\alias{fastqTemplate}
\title{Extract template reads}
\usage{
fastqTemplate(x)
}
\arguments{
\item{x}{Object of class \code{\linkS4class{Fast5Summary}}}
}
\value{
A ShortReadQ object
}
\description{
This generic function accesses the fastq slot stored in an object derived 
from the Fast5Summary class, and returns only the template reads.
}
\examples{
if( require(minionSummaryData) ) {
   data(s.typhi.rep2, package = 'minionSummaryData')
   fastqTemplate( s.typhi.rep2 )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting_summaryStats.R
\name{plotBaseProductionRate}
\alias{plotBaseProductionRate}
\title{Plot the mean rate at which bases are recorded}
\usage{
plotBaseProductionRate(summaryData)
}
\arguments{
\item{summaryData}{Object of class \linkS4class{Fast5Summary}.}
}
\value{
Returns an object of class \code{gg} representing the plot.
}
\description{
For each read, the ratio between the total number of bases called in the read
(template and complement strand, but not 2D composite) and the time spent in
the pore is calculated.  This is then plotted against the time the read 
entered the pore, allow us to assess whether the rate at which callable 
bases are read changes during the experiment run time.
}
\details{
This is likely very similar to \code{\link{plotEventRate}}, although one may
find that large number of events occur that can not be base called, 
resulting in a difference between these two plots.
}
\examples{
if( require(minionSummaryData) ) {
   data(s.typhi.rep2, package = 'minionSummaryData')
   plotBaseProductionRate( s.typhi.rep2 )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Methods-subsetting.R
\name{fastq2D}
\alias{fastq2D}
\title{Extract 2D reads}
\usage{
fastq2D(x)
}
\arguments{
\item{x}{Object of class \code{\linkS4class{Fast5Summary}}}
}
\value{
A ShortReadQ object
}
\description{
This generic function accesses the fastq slot stored in an object derived 
from the Fast5Summary class, and returns only the 2D reads.
}
\examples{
if( require(minionSummaryData) ) {
   data(s.typhi.rep2, package = 'minionSummaryData')
   fastq2D( s.typhi.rep2 )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting_summaryStats.R
\name{plotActiveChannels}
\alias{plotActiveChannels}
\title{Plot the number of active channels for each minute of run time}
\usage{
plotActiveChannels(summaryData)
}
\arguments{
\item{summaryData}{Object of class \linkS4class{Fast5Summary}.}
}
\value{
Returns an object of class \code{gg} representing the plot.
}
\description{
Plot the number of active channels for each minute of run time
}
\examples{
if( require(minionSummaryData) ) {
   data(s.typhi.rep2, package = 'minionSummaryData')
   plotActiveChannels( s.typhi.rep2 )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Methods-accessors.R
\name{fastq}
\alias{fastq}
\title{Extract fastq slot}
\usage{
fastq(x)
}
\arguments{
\item{x}{Object of class \code{\linkS4class{Fast5Summary}}}
}
\value{
A ShortReadQ object
}
\description{
This generic function accesses the fastq slot stored in an object 
derived from the Fast5Summary class.
}
\examples{
if( require(minionSummaryData) ) {
   data(s.typhi.rep2, package = 'minionSummaryData')
   fastq( s.typhi.rep2 )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Methods-accessors.R
\name{readInfo}
\alias{readInfo}
\title{Extract readInfo slot}
\usage{
readInfo(x)
}
\arguments{
\item{x}{Object of class \code{\linkS4class{Fast5Summary}}}
}
\value{
A data.frame with 5 columns
}
\description{
This generic function accesses the readInfo slot stored in an object 
derived from the Fast5Summary class.
}
\examples{
if( require(minionSummaryData) ) {
   data(s.typhi.rep2, package = 'minionSummaryData')
   readInfo( s.typhi.rep2 )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting_summaryStats.R
\name{plotReadCategoryQuals}
\alias{plotReadCategoryQuals}
\title{Visualise the mean base quality of each read}
\usage{
plotReadCategoryQuals(summaryData)
}
\arguments{
\item{summaryData}{Object of class \linkS4class{Fast5Summary}.}
}
\value{
Returns an object of class \code{gg} representing the plot.
}
\description{
Generates a box plot showing the mean base quality for each read, broken down into the three categories of read type that can be found in a fast5 file.
}
\examples{
if( require(minionSummaryData) ) {
   data(s.typhi.rep2, package = 'minionSummaryData')
   plotReadCategoryQuals( s.typhi.rep2 )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readSummary.R
\name{readFast5Summary}
\alias{readFast5Summary}
\title{Read summary data from fast5 files.}
\usage{
readFast5Summary(files)
}
\arguments{
\item{files}{Character vector of fast5 files to be read.}
}
\value{
Object of class \linkS4class{Fast5Summary}
}
\description{
Reads one or more fast5 files and collects summary information about them.
}
\details{
Currently this function assumes all files passed to it come from the same 
sequencing run.  It makes no effort to check for alternative file names or 
the like.  If files from multiple runs are passed to it they will be 
collated together and any analysis performed on them will represent the 
mixture of both experiments.
}
\examples{
\dontrun{
fast5files <- list.files('/foo/bar/', pattern = '.fast5$')
summaryData <- readFast5Summary(fast5files)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fastqProcessing.R
\name{fast5toFastq}
\alias{fast5toFastq}
\title{Extract FASTQ files from fast5 files}
\usage{
fast5toFastq(files, strand = "all", fileName = NULL, outputDir = NULL,
  ncores = 1)
}
\arguments{
\item{files}{Character vector of fast5 files to be read.}

\item{strand}{Character vector specifying the strand to extract.  Can take
any combination of the following options: "template", "complement", "2D", 
"all", "both".}

\item{fileName}{Stem for the name of the names of the output file names.
The appropriate strand will be appended to each file e.g. 
fileName_complement.fq.gz or fileName_template.fq.gz}

\item{outputDir}{Directory output files should be written to.}

\item{ncores}{Specify the number of CPU cores that should be used to process 
the files.  Currently this seems to be more IO bound than CPU, so there
is little benefit achieved by using a high number of cores.}
}
\value{
No value returned.  Run for the side effect of writing the FASTQ
files to disk.
}
\description{
This function provides direct access to the FASTQ entries held within fast5
files.  If you are only interested in getting hold of the base called reads,
and don't require any raw-signal or event information, use this function.
Given a vector of fast5 files, the FASTQ entries will be combined and up to
three gzip compressed FASTQ will be created - one for each of the template, 
complement and 2D strands depending upon what is available in the input 
files.
}
\examples{
\dontrun{
fast5files <- list.files('/foo/bar/', pattern = '.fast5$')
summaryData <- readFast5Summary(fast5files)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fast5Readers.R
\name{readFast5Log}
\alias{readFast5Log}
\title{Read the log entry from a single fast5 file}
\usage{
readFast5Log(file)
}
\arguments{
\item{file}{Character vector of fast5 file to be read.}
}
\value{
Character vector containing the log information.  
\code{NULL} if no log is found.
}
\description{
Basecalling procedures performed on fast5 files generally leave a 
log file entry recording how far through the pipeline the file 
proceeded.  This function will extract this information as a 
single string.  It can be printed in a more readable format 
using the \code{\link{cat}} function.
}
\examples{
fast5file <- system.file('extdata', 'example.fast5', package = "IONiseR")
log <- readFast5Log(fast5file)
cat(log)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Methods-accessors.R
\name{eventData}
\alias{eventData}
\title{Extract eventData slot}
\usage{
eventData(x)
}
\arguments{
\item{x}{Object of class \code{\linkS4class{Fast5Summary}}}
}
\value{
A data.frame with 5 columns
}
\description{
This generic function accesses the eventData slot stored in an object derived 
from the Fast5Summary class.
}
\examples{
if( require(minionSummaryData) ) {
   data(s.typhi.rep2, package = 'minionSummaryData')
   eventData( s.typhi.rep2 )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting_summaryStats.R
\name{plotEventRate}
\alias{plotEventRate}
\title{Plot the mean rate at which events occur}
\usage{
plotEventRate(summaryData)
}
\arguments{
\item{summaryData}{Object of class \linkS4class{Fast5Summary}.}
}
\value{
Returns an object of class \code{gg} representing the plot.
}
\description{
For each read, the ratio between the number of events comprising the read and the time spent in the pore is calculated.  This is then plotted against the time the read entered the pore, allow us to assess whether the rate at which events occur changes during the experiment run time.
}
\examples{
if( require(minionSummaryData) ) {
   data(s.typhi.rep2, package = 'minionSummaryData')
   plotEventRate( s.typhi.rep2 )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R, R/Methods-accessors.R,
%   R/Methods-subsetting.R
\docType{class}
\name{Fast5Summary-class}
\alias{Fast5Summary-class}
\alias{length,Fast5Summary-method}
\alias{readInfo,Fast5Summary-method}
\alias{eventData,Fast5Summary-method}
\alias{baseCalled,Fast5Summary-method}
\alias{fastq,Fast5Summary-method}
\alias{[,Fast5Summary,ANY,ANY,ANY-method}
\alias{fastqTemplate,Fast5Summary-method}
\alias{fastqComplement,Fast5Summary-method}
\alias{fastq2D,Fast5Summary-method}
\title{An S4 class for summarised data from a MinION sequencing run}
\usage{
\S4method{length}{Fast5Summary}(x)

\S4method{readInfo}{Fast5Summary}(x)

\S4method{eventData}{Fast5Summary}(x)

\S4method{baseCalled}{Fast5Summary}(x)

\S4method{fastq}{Fast5Summary}(x)

\S4method{[}{Fast5Summary,ANY,ANY,ANY}(x, i)

\S4method{fastqTemplate}{Fast5Summary}(x)

\S4method{fastqComplement}{Fast5Summary}(x)

\S4method{fastq2D}{Fast5Summary}(x)
}
\arguments{
\item{x}{Object of class Fast5Summary}

\item{i}{Vector defining index to subset by.}
}
\value{
An object of class Fast5Summary
}
\description{
An S4 class for summarised data from a MinION sequencing run
}
\section{Methods (by generic)}{
\itemize{
\item \code{length}: Returns the number of files read during creation of the object

\item \code{readInfo}: Returns readInfo data.frame

\item \code{eventData}: Returns eventData data.frame

\item \code{baseCalled}: Returns baseCalled data.frame

\item \code{fastq}: Returns ShortReadQ object stored in fastq slot.

\item \code{[}: Subset object and return an object of the same class.

\item \code{fastqTemplate}: Returns ShortReadQ object containing only template reads

\item \code{fastqComplement}: Returns ShortReadQ object containing only complement reads

\item \code{fastq2D}: Returns ShortReadQ object containing only 2D reads
}}

\section{Slots}{

\describe{
\item{\code{readInfo}}{Object of class tibble.  Contains five columns: 
\itemize{
  \item id - an integer key that allows use to match entries in the separate 
  slots of this object.
  \item file - Basename of the fast5 file the data was read from.
  \item read - Read number from channel.
  \item channel - channel.
  \item mux - Specific pore that was used within the four that are assigned 
  to a single channel. Should be in the range 1-4, but if this isn't 
  available it will be 0.
}}

\item{\code{rawData}}{Object of class tibble.  Intended to hold raw signal data
although reading this is currently not implemented in \code{IONiseR}.}

\item{\code{eventData}}{Object of class tibble.  Holds summary of events data 
prior to base calling. Contains five columns: 
\itemize{
  \item id - an integer key that allows use to match entries in the 
  separate slots of this object.
  \item start_time - time in seconds after the run started that this 
  reading began. 
  \item duration - time in seconds the reading lasted. 
  \item num_events - the number of events that were recorded as part of 
  this reading. 
  \item median_signal - median of the recorded signals for this set of events.
}}

\item{\code{baseCalled}}{Object of class tibble.  For the most part contains 
similar data to the @eventData slot, the base called data is derived from it.
\itemize{
  \item id - an integer key that allows use to match entries in the 
  separate slots of this object.
  \item start_time - time in seconds after the run started that this 
  reading began. 
  \item duration - time in seconds the reading lasted. 
  \item num_events - the number of events that were recorded as part of 
  this reading. 
  \item strand - can be either 'template' or 'complement'
  \item full_2D - boolean value specifying whether the read forms part of a 
  2D pair.  If TRUE the FASTQ data for the template, complement and 2D read 
  will be available in the @fastq slot.
}}

\item{\code{fastq}}{Object of class ShortReadQ.  This slot contains all reads 
(template, complement and 2D).  The read names take the form NUM_STRAND, 
where NUM matches with the id column in the other slots and STRAND indicates 
whether the read is template, complement or 2D.}

\item{\code{versions}}{A list intended to store the version of IONiseR that was used
to create the object.  (May be extended in the future to include the version
of MinKNOW that the original fast5 files were processed, if this can be 
determined accurately.)}
}}

\examples{
if( require(minionSummaryData) ) {
   data(s.typhi.rep2, package = 'minionSummaryData')
   length( s.typhi.rep2 )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting_summaryStats.R
\name{plotReadAccumulation}
\alias{plotReadAccumulation}
\title{Plot the accumulation of reads over the duration of the experiment.}
\usage{
plotReadAccumulation(summaryData)
}
\arguments{
\item{summaryData}{Object of class \linkS4class{Fast5Summary}.}
}
\value{
Returns an object of class \code{gg} representing the plot.
}
\description{
Plot the accumulation of reads over the duration of the experiment.
}
\examples{
if( require(minionSummaryData) ) {
   data(s.typhi.rep2, package = 'minionSummaryData')
   plotReadAccumulation( s.typhi.rep2 )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting_layout.R
\name{layoutPlot}
\alias{layoutPlot}
\title{Create layout plot of flowcell}
\usage{
layoutPlot(summaryData, attribute = NULL)
}
\arguments{
\item{summaryData}{Object of class \linkS4class{Fast5Summary}.}

\item{attribute}{Character string indicating what to plot. Currently accepted values are: "nreads", "kb", "signal".}
}
\value{
Returns an object of \code{gg} representing the plot.
}
\description{
Creates a plot representing the layout of a MinION flow cell.  Each circle represents an idividual channel with the intensity relecting the total kilobases of sequence produced.  This only considers reads marked as template or complement, 2D reads are ignored as they are generated from the former two.
}
\examples{
if( require(minionSummaryData) ) {
   data(s.typhi.rep2, package = 'minionSummaryData')
   layoutPlot( s.typhi.rep2, attribute = 'nreads' )
   layoutPlot( s.typhi.rep2, attribute = 'kb' )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting_summaryStats.R
\name{channelActivityPlot}
\alias{channelActivityPlot}
\title{Visualise a specified metric over all channels over time.}
\usage{
channelActivityPlot(summaryData, zScale = NULL, zAverage = TRUE)
}
\arguments{
\item{summaryData}{Object of class \linkS4class{Fast5Summary}.}

\item{zScale}{A data.frame containg two columns.  The first must be labelled
'id' and correspond to id field present in all slots in summaryData.  The 
second column should contain data pertaining to that reads that you wish to
be represented on the coloured z-axis.}

\item{zAverage}{Logical indicating if a bar showing the mean across all 
channel for the choosen zScale should be shown on the plot. Defaults to
TRUE.}
}
\value{
Returns an object of class \code{gg} representing the plot.
}
\description{
Plots a line for each fast 5 file, arranged by channel and experiment time 
when the signal was being recorded.  The colour of each line can be 
specified by the user to reflect any metric they wish.  The intention of the
plot is to investigate trends that may appear at specific time points, or 
influence a subset of channels.
}
\examples{
if( require(minionSummaryData) ) {
   require(dplyr)
   data(s.typhi.rep3, package = 'minionSummaryData')
   ## we will plot the median event signal for each read on z-axis
   z_scale = select(eventData(s.typhi.rep3), id, median_signal)
   channelActivityPlot( s.typhi.rep3, zScale = z_scale )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IONiseR.R
\docType{package}
\name{IONiseR}
\alias{IONiseR}
\alias{IONiseR-package}
\title{IONiseR: A package for assessing quality of MinION data}
\description{
IONiseR provides tools for the quality assessment of Oxford Nanopore MinION 
data. It extracts summary statistics from a set of fast5 files and can be 
used either before or after base calling.  In addition to standard summaries
of the read-types produced, it provides a number of plots for visualising 
metrics relative to experiment run time or spatially over the surface of a 
flowcell.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting_kmers.R
\name{plotKmerFrequencyCorrelation}
\alias{plotKmerFrequencyCorrelation}
\title{Display correlation between pentemer proportions in two time windows}
\usage{
plotKmerFrequencyCorrelation(summaryData, kmerLength = 5,
  groupedMinutes = 10, only2D = TRUE)
}
\arguments{
\item{summaryData}{Object of class \linkS4class{Fast5Summary}.}

\item{kmerLength}{Specifies the length of kmers to compare. Defaults to 5 
given the current pentamer reading nature of the nanopores.}

\item{groupedMinutes}{Defines how many minutes each grouping of reads spans.}

\item{only2D}{Logical. If TRUE kmers are computed for only full 2D reads.
If FALSE 2D reads are ignored and all available template and complement 
strands are used.}
}
\value{
Returns an object of class \code{gg} representing the plot.
}
\description{
Plots
}
\examples{
if( require(minionSummaryData) ) {
   data(s.typhi.rep3, package = 'minionSummaryData')
   plotKmerFrequencyCorrelation( s.typhi.rep3, only2D = FALSE )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting_summaryStats.R
\name{plotReadCategoryCounts}
\alias{plotReadCategoryCounts}
\title{Plot the proportion of template, complement and 2D reads found a dataset.}
\usage{
plotReadCategoryCounts(summaryData)
}
\arguments{
\item{summaryData}{Object of class \linkS4class{Fast5Summary}.}
}
\value{
Returns an object of class \code{gg} representing the plot.
}
\description{
Generates a bar plot showing the breakdown of read types found in a set of fast5 files.  There is a strict hierarchy to the types of read that can be found in a fast5 file.  A full 2D read requires both a complement and template strand to have been read correctly.  Similarly, a complement strand can only be present if the template was read successfully.  Finally, you can encounter a file containing now called bases on either strand.
Here we visualise the total number of fast5 files, along with the counts containing each of the categories above.  For an ideal dataset all four bars will be the same height.  This is unlikely, but the drop between bars can give some indication of data quality.
}
\examples{
if( require(minionSummaryData) ) {
   data(s.typhi.rep2, package = 'minionSummaryData')
   plotReadCategoryCounts( s.typhi.rep2 )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Methods-accessors.R
\name{baseCalled}
\alias{baseCalled}
\title{Extract baseCalled slot}
\usage{
baseCalled(x)
}
\arguments{
\item{x}{Object of class \code{\linkS4class{Fast5Summary}}}
}
\value{
A data.frame with 6 columns
}
\description{
This generic function accesses the baseCalled slot stored in an object 
derived from the Fast5Summary class.
}
\examples{
if( require(minionSummaryData) ) {
   data(s.typhi.rep2, package = 'minionSummaryData')
   baseCalled( s.typhi.rep2 )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting_layout.R
\name{channelHeatmap}
\alias{channelHeatmap}
\title{Create layout plot of flowcell}
\usage{
channelHeatmap(data, zValue)
}
\arguments{
\item{data}{A data.frame.  Should have at least two columns, 
one of which has the name 'channel'.}

\item{zValue}{Character string specifying the name of the column to be used 
for the colour scaling.}
}
\value{
Returns an object of \code{gg} representing the plot.
}
\description{
Creates a plot representing the layout of a MinION flow cell.  Each circle
represents an individual channel with the intensity relecting a specified 
sequencing metric.  This function is a more generalised version of 
\code{\link{layoutPlot}}, allowing the user to map any value the like on 
the channel layout.
}
\examples{
library(dplyr)
if( require(minionSummaryData) ) {
   data(s.typhi.rep2, package = 'minionSummaryData')
   ## calculate and plot the mean number of events recorded by each channel
   avgEvents <- left_join(readInfo(s.typhi.rep2), eventData(s.typhi.rep2), by = 'id') \%>\% 
   group_by(channel) \%>\% 
   summarise(mean_nevents = mean(num_events))
   channelHeatmap(avgEvents, zValue = 'mean_nevents')
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting_summaryStats.R
\name{plotCurrentByTime}
\alias{plotCurrentByTime}
\title{View changes in signal against run time.}
\usage{
plotCurrentByTime(summaryData)
}
\arguments{
\item{summaryData}{Object of class \linkS4class{Fast5Summary}.}
}
\value{
Returns an object of class \code{gg} representing the plot.
}
\description{
Plots the median recorded current for each fast5 file against the time at 
which the recording began.
}
\examples{
if( require(minionSummaryData) ) {
   data(s.typhi.rep2, package = 'minionSummaryData')
   plotCurrentByTime( s.typhi.rep2 )
}
}
